"""
test_service.py

Generic test for services (training or prediction). Set up the service folder
(copy model code, download model artifact if necessary), install model
dependencies and set current working directory to service folder, then run:
python -m akerbp.mlops.services.test_service
"""

import json
import subprocess
import sys
import time
import uuid
from importlib import import_module
from pathlib import Path
from typing import Any, Dict, Tuple, Union

from akerbp.mlops.cdf.helpers import get_client
from akerbp.mlops.core import config
from akerbp.mlops.core.helpers import subprocess_wrapper
from akerbp.mlops.core.logger import get_logger

service_name = config.ENV_VARS.service_name
platform = config.ENV_VARS.platform

logger = get_logger(__name__)

client_secrets = config.client_secrets


def mock_saver(*args, **kwargs):
    pass


def run_tests(test_path: Union[Path, str], path_type: str = "file") -> None:
    """
    Run tests with pytest. Raises exception subprocess.CalledProcessError
    if pytest fails.

    Input
      - test_path: path to tests with pytest (string or a list of strings) All
        should have the same format (see next parameter)
      - path_type: either 'file' (test_path refers then to files/folders) or
        'module' (test_path refers then to modules)

    """
    command = [
        sys.executable,
        "-m",
        "pytest",
        "-v",
        "-rA",
        "-o",
        "log_cli=true",
        "--color=no",
        "-W ignore:numpy.ufunc size changed",
    ]
    if path_type == "module":
        command.append("--pyargs")
    if isinstance(test_path, str) or isinstance(test_path, Path):
        command.append(str(test_path))
    elif isinstance(test_path, list):
        command += test_path
    else:
        raise ValueError("Input should be string or list of strings")
    logger.info(f"Run tests: {test_path}")
    subprocess_wrapper(command, stderr=subprocess.STDOUT, cwd=Path.cwd())


def get_model_test_data(test_import_path: str) -> Tuple[Dict, bool]:
    """
    Read input and validation function from the model test file
    """
    logger.info(
        f"Getting test data for {service_name} service from {test_import_path=}"
    )
    service_test = import_module(test_import_path).ServiceTest()
    test_input = getattr(service_test, f"{service_name}_input")
    check = getattr(service_test, f"{service_name}_check")
    return test_input, check  # type: ignore


def test_service(input_payload: Dict, check: Any) -> None:
    """
    Generic service test. Call service with the model's test input data and
    validate the output.
    """
    logger.info(f"Test {service_name} service")
    service = import_module(f"akerbp.mlops.services.{service_name}").service

    if service_name == "training":
        response = service(data=input_payload, secrets=client_secrets, saver=mock_saver)
    elif service_name == "prediction":
        response = service(data=input_payload, secrets=client_secrets)
        # Check for predictions uploaded to CDF Files and delete to avoid unneccesary clogging
        if platform == "cdf" and response["prediction_file"] != "":
            import akerbp.mlops.cdf.helpers as mlops_helpers

            mlops_helpers.client_secrets = client_secrets
            mlops_helpers.set_up_cdf_client(context="write")
            cognite_client = mlops_helpers.global_client["write"]
            prediction_file_external_id = response["prediction_file"]
            logger.info(f"Deleting prediction file {prediction_file_external_id}")
            try:
                cognite_client.files.delete(external_id=prediction_file_external_id)
            except Exception as e:
                logger.warning(f"Error deleting prediction file: {e}")

    else:
        raise Exception("Unknown service name")

    assert response["status"] == "ok"


if __name__ == "__main__":
    logger = get_logger(name="akerbp.mlops.services.test_service.py")

    c = config.read_service_settings()

    # run unit tests
    run_tests(c.test_file)
    test_input, check = get_model_test_data(c.test_import_path)

    # Upload payloads to cdf files, and pass that
    client = get_client(
        client_id=client_secrets["id-write"],
        client_secret=client_secrets["secret-write"],
        tenant_id=client_secrets["tenant-id"],
        base_url=client_secrets["base-url"],
    )
    for i in range(len(test_input["input"])):
        content_to_upload = str(json.dumps(test_input["input"][i]["input_logs"]))
        file_id = uuid.uuid4().hex
        external_file_id = f"mlops_test_service_input_{file_id}.binary"
        file_metadata = client.files.upload_bytes(
            content=content_to_upload,
            name=external_file_id.split(".")[0],
            external_id=external_file_id,
        )
        test_input["input"][i]["input_file_reference"] = file_metadata.external_id
        test_input["input"][i]["input_logs"] = {}
    logger.info(
        "Uploaded input to CDF Files. Testing service with input as a CDF file reference"
    )
    time.sleep(5)  # Give time for CDF to process the file
    test_service(test_input, check)
    # Share input with main process to consume payload for downstream testing of deployed model
    print(json.dumps(test_input))



LOGGING_CONFIG = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "standard": {
            "format": "%(asctime)s - %(name)s - %(levelname)s - %(message)s",
            "datefmt": "%d/%m/%Y %H:%M:%S",
        },
    },
    "handlers": {
        "default": {
            "class": "logging.StreamHandler",
            "level": "INFO",
            "formatter": "standard",
            "stream": "ext://sys.stdout",
        },
    },
    "loggers": {
        "": {"handlers": ["default"], "level": "INFO"},
    },
}



"""
cdf_helpers.py

High level functionality built on top of Cognite's Python SDK

Most functions require the global cdf client to be set up before using them.
"""

import functools
import json
import os
import sys
import time
import traceback
from dataclasses import asdict
from datetime import datetime
from functools import partial
from multiprocessing import Pool
from pathlib import Path
from shutil import make_archive, unpack_archive
from typing import Any, Dict, List, Optional, Tuple, TypedDict, Union, cast

import pandas as pd
from cognite.client import CogniteClient
from cognite.client.config import global_config
from cognite.client.data_classes import FileMetadata, FileMetadataUpdate, Function
from cognite.client.exceptions import CogniteAPIError, CogniteNotFoundError

from akerbp.mlops.core import config, exceptions
from akerbp.mlops.core.helpers import confirm_prompt
from akerbp.mlops.core.logger import get_logger

global_config.disable_pypi_version_check = True
logger = get_logger(__name__)


global_client: Dict[str, CogniteClient] = {}
client_secrets = config.client_secrets

env_vars = config.ENV_VARS


def set_up_cdf_client(context: str = "read") -> None:
    """
    Set up the global client used by most helpers. This needs to be called
    before using any helper.

    Args:
        context (str): either 'read', 'write'. Defaults to 'read'
    """
    client_secret_labels = []
    if context == "read":
        client_secret_labels.append("read")
    elif context == "write":
        client_secret_labels.extend(["read", "write"])
    else:
        raise ValueError("Context should be either 'read', 'write'")

    for k in client_secret_labels:
        if k not in global_client:
            global_client[k] = get_client(
                client_id=client_secrets[f"id-{k}"],
                client_secret=client_secrets[f"secret-{k}"],
                tenant_id=client_secrets["tenant-id"],
                base_url=client_secrets["base-url"],
            )


def get_client(
    client_id: str,
    client_secret: str,
    tenant_id: Optional[str] = None,
    base_url: Optional[str] = None,
) -> CogniteClient:
    """
    Create a CDF client with a given client id and secret

    Args:
        client_id (string): client id for the CDF project
        client_secret (string): client secret for the CDF project
    Returns:
        (CogniteClient): Cognite client to interact with CDF
    """
    from cognite.client import ClientConfig, CogniteClient
    from cognite.client.credentials import OAuthClientCredentials

    # input checks
    if base_url is None:
        base_url = client_secrets["base-url"]
        if base_url is None:
            raise ValueError(
                "base_url is not set. Cannot create a Cognite client without the OIDC base url used in defining scopes."
            )

    if tenant_id is None:
        tenant_id = client_secrets["tenant-id"]
        if tenant_id is None:
            raise ValueError(
                "tenant_id is not set. Cannot create a Cognite client without the tenant id used in defining the token url."
            )

    creds = OAuthClientCredentials(
        token_url=f"https://login.microsoftonline.com/{tenant_id}/oauth2/v2.0/token",
        client_id=client_id,
        client_secret=client_secret,
        scopes=[f"{base_url}/.default"],
    )

    project = "akbp-subsurface"
    cnf = ClientConfig(
        client_name="mlops-client",
        project=project,
        credentials=creds,
        base_url=base_url,
    )

    client = CogniteClient(cnf)
    # assert client.login.status().logged_in
    return client


def validate_function_name(function_name: str, verbose: bool = True) -> bool:
    """
    Validate that function name follows MLOps standard: model-service-env
    Updated to still return true if the external id of the function does
    not follow the mlops convention, to allow listing arbitrary functions.

    Args:
        function_name (string): function name to validate
        verbose (bool, optional): whether to print a warning if the function name is not valid.
            Defaults to True

    Returns:
        (bool): True if name is valid, False otherwise
    """
    supported_services = ["prediction", "training"]
    supported_environments = ["dev", "test", "prod"]
    # Check external id using mlops convention
    try:
        if len(splitted_function_name := function_name.split("-")) > 3:
            _, service, environment, _ = splitted_function_name
        elif len(splitted_function_name) == 3:
            _, service, environment = splitted_function_name
        else:
            _, service, environment = None, None, None
    except ValueError:
        if verbose:
            m = f"Expected function name format: 'model-service-environment', got {function_name}"
            logger.error(m)
        return True
    if service not in supported_services:
        if verbose:
            m = f"Supported services: {supported_services}, got {service} from {function_name}"
            logger.error(m)
        return True
    if environment not in supported_environments:
        if verbose:
            m = f"Supported environments: {supported_environments}, got {environment} from {function_name}"
            logger.error(m)
        return True
    return True


def delete_function(function_name: str, confirm: bool = True) -> None:
    """
    Delete a deployed function

    Args:
        function_name (string): external id of function to delete
        confirm: (bool, optional) whether the user will be asked to confirm deletion.
            Defaults to True
    """
    if not validate_function_name(function_name):
        raise exceptions.FunctionNameError(f"Function name {function_name} is invalid")

    if len(splitted_function_name := function_name.split("-")) > 3:
        model, service, environment, version = splitted_function_name
    else:
        model, service, environment = splitted_function_name

    confirmed = False

    if confirm:
        if len(splitted_function_name) > 3:
            question = f"Delete {model=}, {service=}, {environment=}, {version=}?"
        else:
            question = f"Delete {model=}, {service=}, {environment=}?"
        confirmed = confirm_prompt(question)

    if not confirm or confirmed:
        client = global_client["write"]
        try:
            client.functions.delete(external_id=function_name)
            logger.info(f"Deleted function with external id {function_name}")
        except CogniteNotFoundError:
            logger.error(f"Couldn't find function with external id {function_name}")


def create_function_from_folder(
    human_readable_name: str,
    function_name: str,
    folder: str,
    handler_path: str,
    python_runtime: str,
    description: str = "",
    metadata: Optional[Dict[str, str]] = None,
    owner: str = "",
    secrets: Optional[Dict[str, str]] = None,
) -> Any:
    """
    Create a Cognite function from a folder. Any existing function with the same
    external id is overwritten if it already exist in CDF.

    Args:
        human_readable_name (str): name of function to create
        function_name (str): external id of function to create
        folder (str): path where the source code is located
        handler_path (str): path to the handler file
        python_runtime (str): Python runtime to use for the function
        description (string): model description
        metadata (Dict[str, str], optional): model metadata
            Defaults to an empty dictionary.
        owner (str): the function's owner's email
        secrets (Dict[str, Any], optional): Client secrets or similar that should be passed to the function.
            Defaults to an empty dictionary.

    Returns:
        (Any): the created function
    """
    if secrets is None:
        secrets = {}
    if metadata is None:
        metadata = {}
    client = global_client["write"]

    try:
        client.functions.delete(external_id=function_name)
        logger.info(
            f"Function {human_readable_name} with external id {function_name} already exist and will be overwritten"
        )
    except CogniteNotFoundError:
        logger.info(
            f"Function {human_readable_name} with external id {function_name} does not exist and will be created"
        )
        pass

    envs = {
        **asdict(env_vars),
        "DEPLOYMENT_PLATFORM": "cdf",
    }  # ensure CDFLogger is used
    envs = {k.upper(): str(v) for k, v in envs.items() if v is not None}
    try:
        function = client.functions.create(
            name=human_readable_name,
            folder=folder,
            function_path=handler_path,
            external_id=function_name,
            description=description,
            metadata=metadata,
            runtime=python_runtime,
            owner=owner,
            secrets=secrets,
            env_vars=envs,
        )
        logger.info(f"Environment variables: {envs}")
        logger.info(
            f"Starting deployment of function {human_readable_name} with external id {function_name}"
        )
    except CogniteAPIError:
        logger.info(
            f"Failed to create function {human_readable_name} with external id {function_name}"
        )
        raise

    return function


def create_function_from_file(
    human_readable_name: str,
    function_name: str,
    file_id: int,
    description: str,
    metadata: Dict[str, str],
    owner: str,
    python_runtime: str,
    secrets: Optional[Dict[str, str]] = None,
) -> Any:
    """
    Create a Cognite function from a file deployed to CDF Files.
    If there exist a function with the same external id, the function is
    overwritten.

    Args:
        human_readable_name (str): name of the function to create
        function_name (str): external id of the function to create
        file_id (int): the id for the function file in CDF Files
        description (str): function documentation
        owner (str): the function's owner's email
        python_runtime (str): Python runtime to use for the function
        secrets (dict, optional): Client secrets or similar that should be passed to the function.
            Defaults to an empty dictionary

    Returns:
        (Any): the created function
    """
    if secrets is None:
        secrets = {}
    client = global_client["write"]
    try:
        client.functions.delete(external_id=function_name)
        logger.info(
            f"Function {human_readable_name} with external id {function_name} already exist and will be overwritten"
        )
    except CogniteNotFoundError:
        logger.info(
            f"Function {human_readable_name} with external id {function_name} does not exist and will be created"
        )
        pass

    envs = {
        **asdict(env_vars),
        "DEPLOYMENT_PLATFORM": "cdf",
    }  # ensure CDFLogger is used
    envs = {k.upper(): str(v) for k, v in envs.items() if v is not None}
    try:
        function = client.functions.create(
            name=human_readable_name,
            file_id=file_id,
            external_id=function_name,
            description=description,
            metadata=metadata,
            runtime=python_runtime,
            owner=owner,
            secrets=secrets,
            env_vars=envs,
        )
        logger.info(
            f"Created function {human_readable_name} with external id {function_name}: {file_id=}"
        )
    except CogniteAPIError as e:
        logger.info(
            f"Failed to create function {human_readable_name} with external id {function_name}"
        )
        logger.info(f"Message returned from the Cognite API: {e.message}")

    return function


def robust_create(create_function: partial) -> Function:
    """
    Robust creation of a CDF Function. Wait until the function status is ready
    or failed. If it fails, it will try again `max_error` times

    Args:
        create_function (partial): partial function that creates the CDF function
    """
    max_errors = 3

    for trial in range(max_errors):
        logger.info(f"Creating function (trial {trial + 1}/{max_errors})")
        function = create_function()
        status = wait_function_status(function)
        logger.info(f"Function status is {status}")
        if function.status == "Ready":
            break
        if function.status == "Failed" and trial < max_errors - 1:
            logger.warning(f"Function failed: {function.id=}")
            logger.info(f"Error was: {function.error=}")
            logger.info("Try to create function again")
        else:
            raise Exception(
                f"Function deployment error, exceeded failure limit ({max_errors}): {function.error=}"
            )
    return function


CDFInfo = TypedDict("CDFInfo", {"metadata": dict, "description": str, "owner": str})


def deploy_function(
    human_readable_name: str,
    function_name: str,
    python_runtime: str,
    folder: Optional[str] = None,
    handler_path: Optional[str] = None,
    secrets: Dict[str, str] = client_secrets,
    info: Optional[CDFInfo] = None,
) -> Function:
    """
    Deploys a model as a Cognite function from a folder where the source code is located.
    The argument handler_path points to a file in the folder within which a function named handle is defined.

    Args:
        human_readable_name (str): name of the function to create
        function_name (str): external id of the function to create
        python_runtime (str): the python runtime to use for the function
        folder (str, optional): path where the source code is located.
            Defaults to str(Path().resolve()).
        handler_path (str, optional): path to the handler file.
            Defaults to "handler.py".
        secrets (Dict[str, Any], optional): Client secrets or similar that should be passed to the function.
            Defaults to the global client_secrets dictionary.
        info (Dict[str, Union[str, Dict[str, str]]], optional): dictionary containing info for a specific service, as specified in the settings
            Defaults to a dictionary with empty fields for description, metadata and owner.
    """
    if folder is None:
        folder = str(Path().resolve())
    if handler_path is None:
        handler_path = "handler.py"
    if info is None:
        info = {"description": "", "metadata": {}, "owner": ""}
    try:
        description = info["description"]
    except KeyError:
        raise Exception(
            "Description field is missing, please update mlops_settings.yaml"
        ) from None

    try:
        owner = info["owner"]
    except KeyError:
        raise Exception(
            "Owner field is missing, please update mlops_settings.yaml"
        ) from None

    # Validate metadata one last time
    metadata = info["metadata"]

    # Deploy
    f = functools.partial(
        create_function_from_folder,
        human_readable_name,
        function_name,
        folder,
        handler_path,
        python_runtime,
        description,
        metadata,
        owner,
        secrets,
    )
    return robust_create(f)


def redeploy_function(
    human_readable_name: str,
    function_name: str,
    file_id: int,
    description: str,
    metadata: Dict[str, str],
    owner: str,
    python_runtime: str,
    secrets: Dict[str, str] = client_secrets,
) -> None:
    """
    Deploys a Cognite function from a folder.

    Args:
        function_name (str): name of the function to create
        file_id (int): the id for the function file in CDF Files
        owner (str): the function's owner's email
        description (str): function description
        metadata (Dict[str, str]): dictionary containing function metadata
        python_runtime (str): the python runtime to use for the function
        secrets: (dict, optional) Client secrets or similar that should be passed to the function.
            Defaults to the global client_secrets dictionary.
    """
    f = functools.partial(
        create_function_from_file,
        human_readable_name,
        function_name,
        file_id,
        description,
        metadata,
        owner,
        python_runtime,
        secrets,
    )
    robust_create(f)


def get_function_call_response_metadata(function_id: int) -> Dict[str, Any]:
    """
    Generate metadata for a function
    Args:
        function_id (int): function's id in CDF
    Returns:
        (dict): function call response metadata
    """
    client = global_client["read"]
    function = client.functions.retrieve(id=function_id)

    ts = function.created_time / 1000  # type: ignore
    created_time = datetime.utcfromtimestamp(ts).strftime("%Y-%m-%d %H:%M:%S")

    metadata = {
        "external_id": function.external_id,  # type: ignore
        "description": function.description,  # type: ignore
        "owner": function.owner,  # type: ignore
        "status": function.status,  # type: ignore
        "file_id": function.file_id,  # type: ignore
        "created_time": created_time,
    }
    return metadata


def call_function(function_name: str, data: Dict, retries: int = 1) -> Any:
    """
    Call a function deployed in CDF based on the external id

    Args:
        function_name (str): external id of the deployed function in CDF
        data (dict): model payload

    Returns:
        (dict): function call response
    """
    client = global_client["write"]
    function = client.functions.retrieve(external_id=function_name)
    logger.info(f"Retrieved function with external-id {function_name} from CDF")
    call_complete = False
    t_start = time.process_time()
    while not call_complete:
        try:
            call = function.call(data)  # type: ignore
            call_complete = True
        except CogniteAPIError:
            if retries:
                retries -= 1
                logger.warning(
                    f"Call to {function_name=} failed: Traceback {traceback.format_exc()}"
                )
                logger.warning(f"Retrying function call to {function_name=}")
            else:
                logger.error("Test call to deployed model failed")
                raise
    logger.info(f"Called function ({call.id=})")
    response = call.get_response()
    time.sleep(2)  # Give cognite some time to collect the logs
    logs = call.get_logs()
    for log in logs:
        logger.info(
            "Call log time: %s - Call log message: %s",
            pd.to_datetime(log.timestamp, unit="ms"),
            log.message,
        )
    status = response["status"]  # type: ignore
    if status == "error":
        error_message = response.get("error_message", "No error message provided")  # type: ignore
        exception_message = (
            f"Function call failed with status {status}: {error_message}"
        )
        logger.error(exception_message)
        raise exceptions.FunctionCallError(exception_message)
    if response is None:
        # Something failed in the call. Retrieve the call logs and raise an exception
        logger.error(f"Function call with id {call.id} failed!")
        raise exceptions.MissingResponseError(
            f"Function call with id {call.id} returned an empty response"
        )
    duration = time.process_time() - t_start
    logger.info(f"Function call complete ({call.id=}, {duration=}): {status=}")
    return response


def call_function_process_wrapper(function_name: str, data: Dict) -> Any:
    """
    Set up the cdf client and call a function based on external id.
    This wrapper makes it possible to call a function from an independent processess invoked by the multiprocessing library

    Args:
        function_name (str): function's external id in CDF
        data (dict): data for the call
    Returns:
        (dict): function call response
    """
    set_up_cdf_client("write")
    return call_function(function_name, data)


def call_function_parallel(
    function_name: str, data: Dict, n_calls: Optional[int] = None
) -> Any:
    """
    Make parallel calls to a function deployed in CDF, based on the external id of the function

    Args:
        function_name (str): function's external id in CDF
        data (dict): list of data for the call
        n_calls (int, optional): number of parallel calls (set None to use all available cpu cores).
            Defaults to None.
    """
    f = partial(call_function_process_wrapper, function_name)
    with Pool(n_calls) as p:
        return p.map(f, data)


def test_function(function_name: str, data: Dict) -> None:
    """
    Call a function with data and verify that the response's
    status is 'ok'

    Args:
        function_name (str): function's external id in CDF
        data (dict): data for the call

    Raises:
        ValueError: if the function name is invalid
    """
    logger.info(f"Testing function {function_name}")
    if not validate_function_name(function_name):
        raise ValueError()
    output = call_function(function_name, data)
    assert output["status"] == "ok"
    logger.info("Test call was successful :)")


def wait_function_status(function: Function, status: Optional[List[str]] = None) -> Any:
    """
    Wait until function status is in `status`
    By default it waits for Ready or Failed, which is useful when deploying.
    It implements some control logic, since polling status can fail.

    Args:
        function (Function): function to wait for
        status (List[str], optional): list of statuses to wait for. Defaults to ["Ready", "Failed"].

    Returns:
        (str): function status
    Raises:
        CogniteAPIError: if the function status fails to update
    """
    if status is None:
        status = ["Ready", "Failed"]
    polling_wait_seconds_base = 10.0
    polling_wait_seconds = polling_wait_seconds_base
    max_api_errors_base = 5
    max_api_errors = max_api_errors_base

    logger.info("Wait for function to be ready or to fail")
    while function.status not in status:
        try:
            time.sleep(polling_wait_seconds)
            function.update()
            logger.info(f"{function.status=}")
            polling_wait_seconds = polling_wait_seconds_base
            max_api_errors = max_api_errors_base
        except CogniteAPIError as e:
            max_api_errors -= 1
            logger.warning("Could not update function status, will try again")
            polling_wait_seconds *= 1.2
            if not max_api_errors:
                logger.error("Could not update function status.")
                raise e

    return function.status


def list_functions(tags: Optional[List] = None, model_env: Optional[str] = None) -> Any:
    """
    List deployed functions, optionally filtering by environment (dev, test
    or prod) or set of tags.

    Input:
        - tags: (List[str], optional): list of tags to search for in the function description.
            Defaults to [].
        - model_env: (str, optional) the environment. Defaults to None

    Output:
        - (List[str]): list of function names (i.e. external_id's )
    """
    if tags is None:
        tags = []
    client = global_client["read"]
    functions = client.functions.list(limit=-1)

    def _validate_function(function: Any) -> bool:
        return validate_function_name(function.external_id, verbose=True)

    functions = filter(_validate_function, functions)  # type: ignore
    if model_env is not None:

        def _get_function_environment(function: Any) -> str:
            if function.external_id.split("-") > 1:
                return str(function.external_id.split("-")[2])
            else:
                return "unknown"

        def _env_index(function_name: Any) -> bool:
            return _get_function_environment(function_name) == model_env

        functions = filter(_env_index, functions)  # type: ignore
    if tags:

        def _contains_tag(function: Any, tag: Any) -> bool:
            if tag in function.description:
                return True
            else:
                return False

        def _tags_index(function: Any) -> bool:
            return any(_contains_tag(function, tag) for tag in tags)

        functions = filter(_tags_index, functions)  # type: ignore
    functions = [f.external_id for f in functions]  # type: ignore
    functions = sorted(functions)  # type: ignore
    return functions


def download_file(
    file_id: Dict[str, Union[str, int]], path: Union[Path, str], **kwargs
) -> None:
    """
    Download file from Cognite

    Args:
        id (dict): dictionary with id type (either "id" or "external_id") as key
        path (Union[Path, str]): path to file
    """
    client = global_client["read"]

    external_id = kwargs.get("external_id", None)

    if external_id is not None:
        logger.info(f"Download file with {external_id=} to {path}")
        client.files.download_to_path(path, **external_id)
    else:
        logger.info(f"Download file with {file_id=} to {path}")
        client.files.download_to_path(path, **file_id)  # type: ignore


def upload_file(
    external_id: str,
    path: Union[Path, str],
    metadata: Optional[Dict[str, str]] = None,
    directory: str = "/",
    overwrite: bool = True,
    dataset_id: Optional[int] = None,
) -> FileMetadata:
    """
    Upload file to Cognite

    Args:
        external_id (str): external id
        path (Union[Path, str]): path of local file to upload
        metadata(dict, optional): dictionary with file metadata. Defaults to empty dict.
        directory(str, optional): directory to upload file to. Defaults to "/".
        overwrite (bool, optional): what to do when the external_id exists already. Defaults to True
        dataset_id (int, optional): dataset id. Defaults to None

    Returns:
        (FileMetadata): file metadata
    """
    if metadata is None:
        metadata = {}
    client = global_client["write"]

    metadata = {
        k: v if isinstance(v, str) else json.dumps(v) for k, v in metadata.items()
    }

    file_info = client.files.upload(
        path=path,  # type: ignore
        external_id=external_id,
        metadata=metadata,
        directory=directory,
        overwrite=overwrite,
        data_set_id=dataset_id,  # type: ignore
    )
    return file_info  # type: ignore


def upload_folder(
    external_id: str,
    path: Path,
    metadata: Optional[Dict] = None,
    overwrite: bool = False,
    target_folder: str = "/",
    dataset_id: Optional[int] = None,
) -> FileMetadata:
    """
    Upload folder content to Cognite. It compresses the folder and uploads it.

    Args:
        external_id (str): external id (should be unique in the CDF project)
        path: (Path) path of local folder where content is stored
        metadata (dict, optional): file metadata. Defaults to empty dict
        overwrite (bool, optional): if overwrite==False and `external_id` exists => exception. Defaults to False
        target_folder (str): path where compressed file should be stored. Defaults to "/".
        dataset_id (int, optional): dataset id. Defaults to None

    Returns:
        (FileMetadata): file metadata
    """
    if metadata is None:
        metadata = {}
    fname = external_id.replace("/", "-")
    base_name = path / f"{fname}_archive"
    archive_name = make_archive(str(base_name), "gztar", path)
    file_info = upload_file(
        external_id=external_id,
        path=archive_name,
        metadata=metadata,
        overwrite=overwrite,
        directory=target_folder,
        dataset_id=dataset_id,
    )
    archive_path = Path(archive_name)
    archive_path.unlink()
    return file_info


def download_folder(external_id: str, path: Path) -> None:
    """
    Download content from Cognite to a folder. It is assumed to have been
    uploaded using `upload_folder()`, so it downloads a file and decompresses
    it.

    Args:
        external_id (str): external id
        path: (Path) path of local folder where content will be stored
    """
    fname = external_id.replace("/", "-")
    base_name = path / f"{fname}_archive.tar.gz"
    download_file({"external_id": external_id}, base_name)
    unpack_archive(base_name, base_name.parent)
    base_name.unlink()
    logger.info(f"Model file/s downloaded to {path}")


def log_system_info() -> None:
    """
    Can be called from a handler to log CDF environment information
    """
    logger.debug(f"Python version:\n{os.popen('python --version').read()}")
    logger.debug(f"Python path:\n{sys.path}")
    logger.debug(f"Current working directory:\n{Path.cwd()}")
    logger.debug(f"Content:\n{os.popen('ls -la *').read()}")
    logger.debug(f"Packages:\n{os.popen('pip freeze').read()}")


def query_file_versions(
    directory_prefix: str,
    external_id_prefix: str,
    metadata: Optional[Dict] = None,
    uploaded: Optional[bool] = True,
    dataset_id: Optional[int] = None,
) -> pd.DataFrame:
    """
    Find all file versions that match a query.

    Args:
        directory_prefix (str): directory prefix
        external_id_prefix (str): external id prefix
        metadata (dict): metadata
        uploaded (bool, optional): Whether file has been uploaded. Defaults to True.
        dataset_id (int): dataset id
    Returns:
        (pd.DataFrame): file versions
    """
    if metadata is None:
        metadata = {}
    client = global_client["read"]
    file_list = client.files.list(
        limit=-1,
        directory_prefix=directory_prefix,
        external_id_prefix=external_id_prefix,
        metadata=metadata,
        uploaded=uploaded,  # type: ignore
        data_set_ids=dataset_id,  # type: ignore
    ).to_pandas(camel_case=False)

    return file_list


def delete_file(file_id: Dict) -> None:
    """
    Delete file from Cognite

    Args:
        id (dict): dictionary with id type (either "id" or "external_id") as key
    """
    client = global_client["write"]
    client.files.delete(**file_id)
    logger.info(f"Deleted file with {file_id=}")


def copy_file(
    source_ext_id: str,
    target_ext_id: str,
    overwrite: bool = False,
    dataset_id: Optional[int] = None,
    **kwargs,
) -> None:
    """
    Copy content and metadata of a file in CDF Files

    Args:
        - source_ext_id (str): external id for source file
        - target_ext_id (str): external id for target file
        - overwrite (bool, optional): should target file be overwritten if it exists. Defaults to False
        - dataset_id (int, optional): dataset id for the target file. Defaults to None
    """
    overwrite_name = kwargs.get("overwrite_name", False)
    name = kwargs.get("name", None)
    try:
        client = global_client["write"]
    except KeyError as e:
        raise exceptions.MissingClientError(
            "No CDF client with Files permissions was set up"
        ) from e
    f = client.files.retrieve(external_id=source_ext_id)
    file_content = client.files.download_bytes(external_id=source_ext_id)
    if overwrite_name:
        if name is None:
            raise Exception(
                "You need to set the 'name' kwarg if you want to overwrite the name of the file to copy"
            )
    else:
        name = f.name  # type: ignore
    f = client.files.upload_bytes(
        file_content,
        name=name,
        external_id=target_ext_id,
        metadata=f.metadata,  # type: ignore
        directory=f.directory,  # type: ignore
        overwrite=overwrite,
        data_set_id=dataset_id,  # type: ignore
    )
    m = f"Copied source {source_ext_id} to {target_ext_id}, {f.dump()}"
    logger.info(m)


def file_exists(
    external_id: str, directory: str, dataset_id: Optional[int] = None
) -> bool:
    """
    Check if a file exists in a folder (regardless of uploaded status)

    Args:
        external_id (str): external file-id
        directory (str): directory to search in
        dataset_id (int, optional): dataset id. Defaults to None
    Returns
        (bool): whether file exists
    """
    file_list = query_file_versions(
        directory_prefix=directory,
        external_id_prefix=external_id,
        uploaded=None,
        dataset_id=dataset_id,
    )
    if file_list.empty:
        exists = False
    else:
        exists = True
    return exists


def get_dataset_id(external_id: Optional[str] = None) -> Optional[int]:
    """
    Get dataset id for a file

    Args:
        external_id (str, optional): dataset external id, or `None` for no dataset. Defaults to None
    Returns:
        (int) dataset id, or `None` for no dataset
    """
    if external_id:
        client = global_client["read"]
        dataset = client.data_sets.retrieve(external_id=external_id)
        return int(dataset.id)  # type: ignore
    else:
        return None


def file_to_dataset(
    file_external_id: str, dataset_external_id: Optional[str] = None
) -> None:
    """
    Assign a file to a dataset.

    Args
        file_external_id (str): external id of the file
        dataset_external_id(str, optional): the dataset's external id, `None` for no. Defaults to None
    """
    client = global_client["write"]
    dataset_id = get_dataset_id(dataset_external_id)
    file_metadata_update = FileMetadataUpdate(external_id=file_external_id)
    file_metadata_update = file_metadata_update.data_set_id.set(dataset_id)
    _ = client.files.update(file_metadata_update)
    logger.info(f"Assigned dataset {dataset_id} to file {file_external_id}")


def get_latest_artifact_version(external_id: str) -> int:
    """Extract the latest artifact version based on the external id of a function.
    If no artifacts are found for the specified function, the version is set to 1.

    Args:
        external_id (string): external id of the function

    Returns:
        (int): latest artifact version of the function
    """
    if len(splitted_external_id := external_id.split("-")) > 3:
        env = splitted_external_id[-2]
    else:
        env = splitted_external_id[-1]
    model_name = splitted_external_id[0]

    artifact_external_id = model_name + "/" + env
    artifact_versions = query_file_versions(
        directory_prefix="/mlops", external_id_prefix=artifact_external_id
    )
    if len(artifact_versions) == 0:
        raise ValueError(
            f"No artifacts found for function with external id {external_id}. Please ensure you upload your artifacts before starting deployment."
        )
    else:
        latest_artifact_version = artifact_versions.loc[
            artifact_versions["uploaded_time"].argmax()
        ].metadata["version"]

    return int(latest_artifact_version)


def get_arguments_for_redeploying_numbered_models(
    external_id: str,
) -> Tuple[str, str, str, Dict[str, str], str]:
    """Extract data needed to redeploy the latest function from CDF files using a predictable external id,
    i.e. witout the model version number. In this way the client does not have to care about version numbers.
    In addition to copying the data fields of the function, the version number is added to the metadata (set to 1 if it does not exist)

    Args:
        external_id (str): external id of the deployed function in CDF

    Returns:
        (tuple): tuple containing the data needed for redeploying the latest function with a predictable external id
    """
    client = global_client["read"]
    latest_function = client.functions.retrieve(external_id=external_id)
    name = latest_function.name  # type: ignore
    file_id = latest_function.file_id  # type: ignore
    description = latest_function.description  # type: ignore
    owner = latest_function.owner  # type: ignore
    metadata = latest_function.metadata  # type: ignore
    return name, file_id, description, metadata, owner  # type: ignore


def setup_schedule_for_latest_model_in_prod(
    external_id: str,
    function_id: int,
    client: Optional[CogniteClient] = None,
) -> None:
    if client is None:
        client = global_client["write"]
    schedules = client.functions.schedules.list(
        function_external_id=external_id,
    )
    if len(schedules) > 0:
        logger.info("A schedule already exist and will be overwritten")
        for schedule in schedules:
            schedule_id = cast(int, schedule.id)
            client.functions.schedules.delete(id=schedule_id)

    _ = client.functions.schedules.create(
        name="Keep warm schedule",
        description="Keep the function warm by calling it with an empty payload every 30 minutes during extended working hours on weekdays",
        cron_expression="*/30 5-17 * * 1-5",
        function_id=function_id,
        data={},
    )
    logger.info("Schedule created")


def garbage_collection(
    c: config.ServiceSettings,
    function_name: str,
    env: str,
    client: Optional[CogniteClient] = None,
    remove_artifacts: bool = False,  # for mlops only
) -> None:
    """Garbage collection of old models in CDF"""
    models_to_keep = c.models_to_keep
    keep_all_models = c.keep_all_models if c.keep_all_models is not None else False
    if client is None:
        client = global_client["read"]
    if models_to_keep is None:
        if keep_all_models is True:
            logger.info("Keeping all models alive as specified in the settings file")
            return
        else:
            logger.info("Number of models to keep alive not specified in settings file")
            logger.info("Setting number of models to keep to 3 (default)")
            models_to_keep = 3
    else:
        logger.info(
            f"Number of models to keep alive inferred from settings file ({models_to_keep})"
        )
        models_to_keep = int(models_to_keep)

    existing_functions: List[str] = [
        cast(str, f.external_id)
        for f in client.functions.list(external_id_prefix=function_name)
    ]
    version_ids = [eid for eid in existing_functions if eid.count("-") == 3]
    latest_artifact_version = get_latest_artifact_version(external_id=function_name)

    if (
        models_to_keep is not None
        and models_to_keep >= len(version_ids)
        and len(version_ids) != 0
    ):
        logger.info(
            f"Number of alive models ({len(version_ids)}) fewer than/equal to the number of models to keep alive ({models_to_keep}). Skipping garbage collection."
        )
    elif len(version_ids) == 0:  # Special case for mlopsdemo during testing
        if remove_artifacts:  # equivalent to c.model_name == "mlopsdemo"
            logger.info(
                f"Deleting non-numbered models of model {c.model_name} in {env}"
            )
            delete_function(function_name=function_name, confirm=False)
            m, _, e = function_name.split("-")
            if (
                latest_artifact_version != 1
            ):  # Need at least a single artifact to be able to deploy
                delete_file(
                    file_id={"external_id": f"{m}/{e}/{latest_artifact_version}"}
                )

    else:
        logger.info(
            f"Starting garbage collection of model {c.human_friendly_model_name} in {env}"
        )

        external_ids_sorted = [
            function_name + "-" + str(v)
            for v in sorted([
                int(external_id.split("-")[-1]) for external_id in version_ids
            ])
        ]

        num_models_to_delete = len(version_ids) - models_to_keep
        external_ids_to_delete = external_ids_sorted[:num_models_to_delete]

        for external_id in external_ids_to_delete:
            version = external_id.split("-")[-1]
            logger.info(f"Deleting version {version} of model {c.model_name} in {env}")
            delete_function(function_name=external_id, confirm=False)
            if remove_artifacts is True:
                m, _, e, v = external_id.split("-")
                delete_file(file_id={"external_id": f"{m}/{e}/{v}"})

        if remove_artifacts:
            # Remove any non versioned models too (artifacts already removed)
            non_versioned_models = set(existing_functions) - set(version_ids)
            for external_id in non_versioned_models:
                logger.info(f"Deleting non-versioned model {external_id} in {env}")
                delete_function(function_name=external_id, confirm=False)


def cdf_metadata_validation(
    metadata: dict,
) -> None:
    metadata_fields = metadata.keys()
    if (no_of_metadata_fields := len(metadata_fields)) > 16:
        # Cognite limitation is 16, but we add model version, mlops version and mlpet version
        # to metadata on deployment.
        raise ValueError(
            f"Too many fields in the metadata specification, exceeding the maximum of 13 fields. Number of fields in settings file {no_of_metadata_fields}"
        )
    try:
        for key, value in metadata.items():
            key_ = str(key)
            value_ = str(value)
            if len(key_) > 32:
                raise ValueError(
                    f"Key '{key}' exceeds the maximum length of 32 characters ({len(key_)})."
                )

            if len(value_) > 512:
                raise ValueError(
                    f"Value '{value}' of '{key}' exceeds the maximum length of 512 characters ({len(value_)})."
                )

        # Maximum size of entire metadata cannot exceed 4096 bytes
        if (metadata_size := len(json.dumps(metadata).encode())) > 4096:
            raise ValueError(
                f"Size of metadata exceeds the maximum size of 4096 bytes ({metadata_size})."
            )
    except ValueError:
        logger.error(f"Metadata validation failed: {metadata=}")
        raise


def delete_test_payload(external_id: str) -> None:
    client = global_client["write"]
    try:
        client.files.delete(external_id=external_id)
        logger.info(f"Deleted test payload with external id {external_id}")
    except CogniteNotFoundError:
        logger.info(f"Test payload with external id {external_id} does not exist")
        pass
    except CogniteAPIError:
        logger.error(f"Failed to delete test payload with external id {external_id}")



# config.py
import os
import re
import traceback
from dataclasses import asdict
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, Union

import yaml
from pydantic import FilePath
from pydantic.dataclasses import dataclass

from akerbp.mlops.core import helpers
from akerbp.mlops.core.exceptions import (
    FieldValueError,
    MissingFieldError,
    MissingMetadataError,
)
from akerbp.mlops.core.logger import get_logger

logger = get_logger(__name__)


def validate_categorical(
    setting: Union[str, None, bool], name: str, allowed: List[Optional[str]]
) -> None:
    """Helper function to validate categorical variables in the framework

    Args:
        setting (Optional[str]): variable to validate. Defaults to None
        name (str): name of the variable to validate
        allowed (List[Optional[str]]): list of allowed values for the variable.

    Raises:
        ValueError: message with the name of the invalid variable and the allowed values
    """
    if setting not in allowed:
        m = f"{name}: allowed values are {allowed}, got '{setting}'"
        raise ValueError(m)


@dataclass
class EnvVar:
    """Dataclass for keeping track of environmental variables

    Attributes:
        model_env (Optional[str]): environment. Defaults to None
        service_name (Optional[str]): service name. Defaults to None
        platform (Optional[str]): platform. Defaults to None
    """

    model_env: Optional[str] = None
    service_name: Optional[str] = None
    platform: Optional[str] = None
    testing_only: Optional[Union[str, bool]] = False

    def __post_init__(self):
        """Post initialization function of the dataclass that validates the attributes env, platform and service_name
        Allowed values for attributes:
            model_env: 'dev', 'test', 'prod'
            platform: 'cdf', 'local'
            service_name: 'training', 'prediction'
        """
        if self.model_env:
            validate_categorical(self.model_env, "Environment", ["dev", "test", "prod"])
        else:
            logger.warning("MODEL_ENV environmental variable is not set")
        validate_categorical(self.platform, "Platform", ["cdf", "local"])
        if self.model_env and self.model_env != "dev":
            validate_categorical(
                self.service_name, "Service  name", ["training", "prediction"]
            )


def read_env_vars() -> EnvVar:
    """
    Read environmental variables and initialize EnvVar object with those that
    were set (i.e. ignored those with None value). Note that
    Default values:
        DEPLOYMENT_PLATFORM: 'cdf'
        TESTING_ONLY: 'False'
    """
    envs = {
        "model_env": os.getenv("MODEL_ENV"),
        "service_name": os.getenv("SERVICE_NAME"),
        "platform": os.getenv("DEPLOYMENT_PLATFORM", "cdf"),
        "testing_only": os.getenv("TESTING_ONLY", "False"),
    }
    envs = {k: v for k, v in envs.items() if v is not None}

    return EnvVar(
        model_env=envs.get("model_env"),
        service_name=envs.get("service_name"),
        platform=envs.get("platform"),
        testing_only=envs.get("testing_only"),
    )


ENV_VARS = read_env_vars()


@dataclass
class AvailableSecrets:
    """
    dataclass for keeping track of secrets available to the model endpoint
    for authentication

    Attributes:
        tenant_id (Optional[str]): CDF tenant id. Defaults to None
        base_url (Optional[str]): CDF base url. Defaults to None
        id_read (Optional[str]): CDF client id for read operations. Defaults to None
        secret_read (Optional[str]): CDF client secret for read operations. Defaults to None
        id_write (Optional[str]): CDF client id for write operations. Defaults to None
        secret_write (Optional[str]): CDF client secret for write operations. Defaults to None
    """

    tenant_id: Optional[str] = None
    base_url: Optional[str] = None
    id_read: Optional[str] = None
    secret_read: Optional[str] = None
    id_write: Optional[str] = None
    secret_write: Optional[str] = None
    osdu_id: Optional[str] = None
    osdu_secret: Optional[str] = None
    osdu_base_url: Optional[str] = None
    osdu_dp_id: Optional[str] = None

    def __post_init__(self):
        """Post initialization function of the dataclass that validates the attribute base_url"""
        # Ensure there are no trailing slashes in the base url
        if self.base_url is not None:
            self.base_url = re.sub(r"/$", "", self.base_url)

        if self.osdu_base_url is not None:
            self.osdu_base_url = re.sub(r"/$", "", self.osdu_base_url)


def read_secrets() -> Dict[str, str]:
    secrets = asdict(
        AvailableSecrets(
            tenant_id=os.getenv("COGNITE_TENANT_ID"),
            base_url=os.getenv("COGNITE_OIDC_BASE_URL"),
            id_read=os.getenv("COGNITE_CLIENT_ID_READ"),
            secret_read=os.getenv("COGNITE_CLIENT_SECRET_READ"),
            id_write=os.getenv("COGNITE_CLIENT_ID_WRITE"),
            secret_write=os.getenv("COGNITE_CLIENT_SECRET_WRITE"),
            osdu_id=os.getenv("OSDU_CLIENT_ID"),
            osdu_secret=os.getenv("OSDU_CLIENT_SECRET"),
            osdu_base_url=os.getenv("OSDU_API_BASE_URL"),
            osdu_dp_id=os.getenv("OSDU_DATA_PARTITION_ID"),
        )
    )
    return {k.replace("_", "-"): v for k, v in secrets.items() if v is not None}


client_secrets = read_secrets()


def generate_default_project_settings(
    yaml_file: Path = Path("mlops_settings.yaml"), n_models: int = 2
) -> None:
    """Generate default the mlops_settings.yaml file from a hardcoded template

    Args:
        yaml_file (Path, optional): path to the mlops_settings.yaml file. Defaults to Path("mlops_settings.yaml")
        n_models (int, optional): number of models to generate. Defaults to 2

    """
    if yaml_file.exists():
        raise Exception(f"Settings file {yaml_file} exists already.")

    default_config_template = [
        """
model_name: my_model
human_friendly_model_name: 'My model'
model_file: model_code/my_model.py
req_file: model_code/requirements.model
test_file: model_code/test_model.py
artifact_folder: artifact_folder
platform: cdf
dataset: mlops
helper_models:
    - model_name: helper_model_1
      env: dev
      version: 1
      target_path: model_code/helper_model_1
packages_to_check:
    - akerbp-mlpet
info:
    prediction:
        description: Description prediction service for my_model
        metadata:
            required_input:
                - input_1
                - input_2
            optional_input:
                - optional_input_1
                - optional_input_2
            keyword_arguments:
                - name: keyword_1
                  required: True
                - name: keyword_2
                  required: False
                  default: 0
            training_wells:
                - 3/1-4
            input_types:
                - float
                - int
            units:
                - s/ft
                - 1
            output_curves:
                - output_1
            output_units:
                - s/ft
            petrel_exposure: False
            supports_external_retrieval: False
            imputed: True
            num_filler: -999.15
            cat_filler: UNKNOWN
        owner: datascientist@akerbp.com
"""
    ]
    default_config_list = default_config_template * n_models
    default_config = "---".join(default_config_list)
    with yaml_file.open("w") as f:
        f.write(default_config)


def validate_model_reqs(req_file: FilePath) -> None:
    # Model reqs is renamed to requirements.txt during deployment
    if req_file.name == "requirements.model":
        with req_file.open() as f:
            req_file_string = f.read()
            if "akerbp-mlops" not in req_file_string:
                m = "Model requirements should include akerbp-mlops package"
                raise Exception(m)
            if "MLOPS_VERSION" not in req_file_string:
                m = 'akerbp-mlops version should be "MLOPS_VERSION"'
                raise Exception(m)


@dataclass
class HelperModel:
    """dataclass for keeping track of helper models

    Attributes:
        model_name (str): name of the helper model
        env (Optional[str]): environment of the helper model. Defaults to None
        version (Optional[int]): version of the helper model. Defaults to None
        target_path (Optional[str]): path to the helper model. Defaults to None
    """

    model_name: str
    env: Optional[str] = None
    version: Optional[int] = None
    target_path: Optional[str] = None


@dataclass
class ServiceSettings:
    """dataclass for keeping track of service settings variables

    Attributes:
        model_name (str): name of the model
        human_friendly_model_name (str): human friendly name of the model, displayed on CDF
        model_file (FilePath): path to the model interface
        req_file (FilePath): path to the model requirements
        info (dict): info of the model. Defaults to None
        test_file (FilePath, optional): path to the model test suite. Defaults to None
        artifact_folder (FilePath, optional): path to the model artifacts. Defaults to None
        platform (str, optional): platform of the model. Defaults to "cdf"
        dataset (str, optional): dataset of the model. Defaults to "mlops"
        model_id (str, optional): id of the model. Defaults to None

    """

    model_name: str  # Remember to modify generate_default_project_settings()
    human_friendly_model_name: str
    model_file: FilePath  # if fields are modified
    req_file: FilePath
    info: Dict
    test_file: Optional[FilePath] = None
    artifact_folder: Optional[Path] = None
    artifact_version: Optional[int] = None
    keep_all_models: Optional[bool] = None
    models_to_keep: Optional[int] = None
    helper_models: Optional[List[HelperModel]] = None
    packages_to_check: Optional[List[str]] = None
    python_version: str = "not_set"
    platform: str = "cdf"
    dataset: str = "mlops"
    model_id: Optional[str] = None

    def __post_init_post_parse__(self):
        """post initialization function of the dataclass that validates model requirements and deployment platform variables.
        Moreover, the model interface and test suite import paths are set as class attributes,
        as well as a dictionary containing the required files for running the model (based on deployment platform)

        Raises:
            Exception: If model name contains special characters not supported by the framework
            ValueError: If specifying a model id when deploying a training service
        """
        envs = ENV_VARS
        # Validation
        if not re.match("^[A-Za-z0-9_]*$", self.model_name):
            m = "Model name can only contain letters, numbers and underscores"
            raise Exception(m)

        validate_model_reqs(self.req_file)

        validate_categorical(self.platform, "Deployment platform", ["cdf", "local"])

        if self.platform == "cdf":
            validate_python_runtime(self.python_version)

        if self.model_id and envs.service_name == "training":
            raise ValueError("Unexpected model_id setting (training service)")

        self.model_import_path = helpers.as_import_path(self.model_file)
        self.test_import_path = helpers.as_import_path(self.test_file)

        self.files = {
            "model code": helpers.get_top_folder(self.model_file),
            "handler": ("akerbp.mlops.cdf", "handler.py"),
            "artifact folder": self.artifact_folder,
        }


def store_service_settings(
    c: ServiceSettings, yaml_file: Path = Path("mlops_service_settings.yaml")
) -> None:
    """Store service settings in a yaml file during deployment

    Args:
        c (ServiceSettings): service settings for the model to deploy
        yaml_file (Path, optional): path to service settings. Defaults to Path("mlops_service_settings.yaml").

    """
    logger.info("Write service settings file to %s", yaml_file.resolve())

    def factory(data: List[Tuple[str, Any]]) -> Dict[str, str]:
        """
        Take a list of tuples as input. Returns a suitable dictionary.
        Transforms Path objects to strings (linux style path).

        Args:
            data (List[Tuple[str, Any]]): list of tuples to transform

        Returns:
            Dict[str, str]: dictionary with the transformed paths
        """

        def path2str(x: Union[Path, str]) -> str:
            """transforms a Path to a string

            Args:
                x (Union[Path, str]): input path

            Returns:
                str: output string
            """
            if not isinstance(x, Path):
                return x
            else:
                return x.as_posix()

        d = {k: path2str(v) for k, v in data}
        return d

    service_settings = asdict(c, dict_factory=factory)
    with yaml_file.open("w") as f:
        yaml.dump(service_settings, f)


@dataclass
class ProjectSettings:
    """
    dataclass for keeping track of project settings

    Attributes:
        project_settings (List[ServiceSettings]): list of service settings
    """

    project_settings: List[ServiceSettings]


def enforce_string_values_in_function_metadata(
    project_settings: ProjectSettings,
) -> ProjectSettings:
    """The metadata field in CDF functions requires both keys and values to be strings.
    This function iterates through the metadata of each model defined in the mlops settings file,
    and enforce the values to be string (keys are strings by default)

    Args:
        project_settings (ProjectSettings): project settings for each model defined in the settings file

    Returns:
        (ProjectSettings): project settings for each model defined in the settings file
    """
    for i, model_settings in enumerate(project_settings.project_settings):
        for service in list(model_settings.info.keys()):
            metadata = model_settings.info[service]["metadata"]
            for k, v in metadata.items():
                metadata[k] = str(v)
            model_settings.info[service]["metadata"] = metadata
        project_settings.project_settings[i] = model_settings
    return project_settings


def read_project_settings(
    yaml_file: Path = Path("mlops_settings.yaml"),
) -> List[ServiceSettings]:
    """Read project settings from the mlops_setting.yaml file

    Args:
        yaml_file (Path, optional): path to mlops settings. Defaults to Path("mlops_settings.yaml").

    Returns:
        List[ServiceSettings]: list of service settings for each model specified in the settings file
    """
    logger.info("Read project settings")
    with yaml_file.open() as f:
        settings = yaml.safe_load_all(f.read())

    downstream_settings = []
    for setting in settings:
        downstream_settings.append(setting)
        valid_metadata, missing_fields = settings_validation(settings=setting)
        if valid_metadata:
            continue
        else:
            logger.error(
                "Invalid metadata specification due to missing fields or invalid values"
            )
            error_message = f"The following field(s) are missing from the metadata specification: {missing_fields}"
            raise MissingMetadataError(error_message)

    model_settings = [ServiceSettings(**s) for s in downstream_settings]
    project_settings = ProjectSettings(project_settings=model_settings)
    project_settings = enforce_string_values_in_function_metadata(project_settings)

    return project_settings.project_settings


def settings_validation(
    settings: dict,
) -> Tuple[bool, List[str]]:
    """Checks whether the mlopds settings contains all the required fields, as specified on confluence

    Args:
        settings (dict): mlops settings as dictionary

    Returns:
        bool: whether metadata is validated
    """
    required_fields = [
        "required_input",
        "optional_input",
        "output_curves",
        "petrel_exposure",
        "keyword_arguments",
        "supports_external_retrieval",
    ]

    required_fields_petrel = [
        "petrel_template_family",
    ]
    metadata = settings["info"]["prediction"]["metadata"]
    metadata_fields = metadata.keys()

    missing_fields = []
    for field in required_fields:
        if field == "petrel_exposure":
            try:
                petrel_exposure = metadata[field]
                if petrel_exposure:
                    for petrel_field in required_fields_petrel:
                        if petrel_field not in metadata_fields:
                            missing_fields.append(petrel_field)
            except KeyError:
                pass

        if field not in metadata_fields:
            missing_fields.append(field)

    if len(missing_fields) == 0:
        return True, missing_fields
    else:
        return False, missing_fields


def validate_python_runtime(python_runtime: str) -> None:
    """Validates the python runtime specified in the mlops settings file

    Args:
        settings (dict): mlops settings as dictionary

    Raises:
        ValueError: if python runtime is not supported
        MissingFieldError: if python runtime is not specified
    """
    # TODO: keep below list updated if cognite sdk adds support for new python runtimes
    allowed_runtimes = ["py37", "py38", "py39", "py310"]
    if python_runtime == "not_set":
        raise MissingFieldError(
            "Python runtime not specified. Please ensure to included it in your mlops_settings.yaml file! "
            f"Supported versions are {allowed_runtimes}."
        )
    elif python_runtime not in allowed_runtimes:
        raise FieldValueError(
            f"Python version {python_runtime} is not supported. Supported versions are {allowed_runtimes}"
        )


def read_service_settings(
    yaml_file: Path = Path("mlops_service_settings.yaml"),
) -> ServiceSettings:
    """Read service settings from the mlops_service_settings.yaml file during deployment

    Args:
        yaml_file (Path, optional): path object to settings file. Defaults to Path("mlops_service_settings.yaml").

    Returns:
        ServiceSettings: service settings for the model to deploy
    """
    logger.info("Read service settings")
    with yaml_file.open() as f:
        settings = yaml.safe_load(f.read())
    service_settings = ServiceSettings(**settings)
    return service_settings


def validate_user_settings(yaml_file: Path = Path("mlops_settings.yaml")) -> None:
    """Validate the mlops_settings.yaml file

    Args:
        yaml_file (Path, optional): path to settings file. Defaults to Path("mlops_settings.yaml").
    """
    try:
        read_project_settings(yaml_file)
        logger.info("Settings file is ok :)")
    except Exception:
        trace = traceback.format_exc()
        error_message = f"Settings file is not ok! Fix this:\n{trace}"
        logger.error(error_message)



"""
deploy.py

Deploy services in either Google Cloud Run or CDF Functions.
Model registry uses CDF Files.
"""

import ast
import os
import traceback
from pathlib import Path
from typing import Dict, List, Optional

import akerbp.mlops.model_manager as mm
from akerbp.mlops import __version__ as package_version
from akerbp.mlops.core.config import (
    HelperModel,
    ServiceSettings,
    read_env_vars,
    read_project_settings,
    store_service_settings,
)
from akerbp.mlops.core.exceptions import DependencyMismatchError

# Global variables
from akerbp.mlops.core.logger import get_logger
from akerbp.mlops.deployment import helpers

logger = get_logger(__name__)


def deploy_model(  # noqa: C901
    model_settings: ServiceSettings,
) -> str:
    """
    Deploy a model.

    This will create a deployment folder and change current working directory
    to it.

    Return "OK" if deployment was successful, otherwise return a string with the traceback for the failed deployment

    Args:
        model_settings (ServiceSettings): settings for the model service
        platform_methods (dict): where key is the platform and value is a tuple with deploy and test functions.
            Defaults to the globally set platform_methods variable.

    Returns:
        (str): status of the deployment
    """
    c = model_settings
    try:
        original_req_file = c.req_file
        envs = read_env_vars()
        env = envs.model_env
        testing_only = eval(envs.testing_only)
        service_name = envs.service_name
        platform = envs.platform
        if service_name is None:
            raise ValueError(
                "SERVICE_NAME environment variable must be set to either 'prediction' or 'training'"
            )
        deployment_folder = helpers.deployment_folder_path(c.model_name)
        function_name = f"{c.model_name}-{service_name}-{env}"

        # Handle mlops internal testing settings
        mlops_testing = ast.literal_eval(os.environ.get("LOCAL_MLOPS_TESTING", "False"))  # type: ignore

        logger.info(
            f"Starting deployment and/or testing of model {c.human_friendly_model_name}"
        )

        if (service_name == "prediction") and c.artifact_folder:
            if platform == "cdf" or platform == "local":
                mm.set_active_dataset(c.dataset)
                c.model_id = mm.set_up_model_artifact(c.artifact_folder, c.model_name)

        logger.info("Create deployment folder and move required files/folders")
        deployment_folder.mkdir()
        # Write settings to deployment folder
        store_service_settings(c, deployment_folder / "mlops_service_settings.yaml")
        logger.info("Service settings %s", c)
        # Update requirements file with mlops version
        req_file = helpers.create_temporary_copy(c.req_file)
        c.req_file = req_file

        helpers.set_mlops_import(
            req_file=c.req_file,
            platform=c.platform,
            deployment_folder=deployment_folder if mlops_testing else None,
        )
        helpers.copy_to_deployment_folder(c.files, deployment_folder)

        logger.info(f"cd {deployment_folder}")
        os.chdir(deployment_folder)

        # Main model requirements
        main_model_requirements = helpers.requirements_to_dict(c.req_file)

        handle_helper_models_and_dependency_mismatch(
            c, main_model_requirements, c.helper_models, c.packages_to_check
        )

        # Deploy to CDF
        if c.platform == "cdf":
            if testing_only:
                logger.info(
                    f"Running tests for model {c.human_friendly_model_name} in {env}, will not deploy"
                )
                helpers.run_tests(c, setup_venv=True, deploy=False)
            else:
                helpers.do_deploy(
                    c,
                    env,
                    service_name,
                    function_name,
                    deployment_folder,
                    setup_venv=True,
                )
        return "OK"
    except (Exception, KeyboardInterrupt):
        # Remove deployment folder if it exists
        helpers.rm_deployment_folder(c.model_name)
        # Remove temporary requirements file if it exists
        if c.req_file != original_req_file:
            req_file = Path(c.req_file).resolve()
            logger.info("Removing temporary requirements file %s", req_file)
            if req_file.exists():
                req_file.unlink()
        trace = traceback.format_exc()
        return f"Model failed to deploy and/or tests failed! See the following traceback for more info: \n\n{trace}"


def handle_helper_models_and_dependency_mismatch(
    c: ServiceSettings,
    main_model_requirements: Dict[str, str],
    helper_models: Optional[List[HelperModel]],
    packages_to_check: Optional[List[str]],
):
    if helper_models is None:
        if packages_to_check is not None:
            logger.warning(
                "packages_to_check specified in mlops_settings, but helper_models is not. Will not check for dependency mismatch between main and helper models."
            )
        return
    root = Path().resolve()
    for helper_model in helper_models:
        logger.info(
            f"Checking for dependency mismatch between main model {c.model_name} and helper model {helper_model.model_name} for the following package(s): {packages_to_check}"
        )
        model_name = helper_model.model_name
        env = helper_model.env if helper_model.env is not None else "prod"
        version = helper_model.version
        helper_model_target = (
            root / helper_model.target_path
            if helper_model.target_path is not None
            else root / model_name
        )
        mm.download_deployment_folder(
            model_name=model_name,
            env=env,
            version=version,
            target_path=str(helper_model_target),
            files_to_ignore=[
                "handler.py",
                "mlops_service_settings.yaml",
            ],
        )
        path_to_helper_model_requirements = helper_model_target / "requirements.txt"
        if packages_to_check is None:
            logger.warning(
                f"{model_name} specified in mlops_settings as a helper model, but packages_to_check is not. Will not check for dependency mismatch between the main and {model_name} models."
            )
        else:
            # Build dictionary of helper model requirements
            helper_model_requirements = helpers.requirements_to_dict(
                path_to_helper_model_requirements
            )
            (
                is_dependency_mismatch,
                mismatched_packages,
            ) = helpers.check_mismatch_in_model_requirements(
                main_model_requirements=main_model_requirements,
                helper_model_requirements=helper_model_requirements,
                packages_to_check=packages_to_check,
            )
            if is_dependency_mismatch:
                raise DependencyMismatchError(
                    f"Dependency mismatch detected for helper model {helper_model} for the following package(s): {mismatched_packages}"
                )
            else:
                logger.info("No dependency mismatch detected for helper model")

        # Remove helper_model requirements file as this interferes with Cognite deployment
        path_to_helper_model_requirements.unlink()


def deploy(project_settings: List[ServiceSettings]) -> None:
    """
    Deploy a machine learning project that potentially contains multiple models.
    Deploy each model in the settings and make sure that if one model fails it
    does not affect the rest. At the end, if any model failed, it raises an
    exception with a summary of all models that failed.

    Args:
        Project settings as described by the user in the config file.

    Raises:
        Exception: If any model failed to deploy.
    """
    failed_models = {}
    cwd_path = Path.cwd()

    for c in project_settings:
        status = deploy_model(c)
        if status != "OK":
            logger.error(status)
            failed_models[c.human_friendly_model_name] = status

        logger.info("cd ..")
        os.chdir(cwd_path)
        helpers.rm_deployment_folder(c.model_name)

    if failed_models:
        for model, message in failed_models.items():
            logger.warning(f"Model {model} failed: {message}")
        raise Exception("At least one model failed.")


def main() -> None:
    logger = get_logger(name="akerbp.mlops.deployment.deploy.py")
    if ast.literal_eval(os.environ.get("LOCAL_MLOPS_TESTING", "False")):
        message = "Deploying prediction service using the built wheel for testing purposes before releasing"
    else:
        message = f"Deploying prediction service using MLOps framework version {package_version}"

    logger.info(message)
    mm.setup()
    settings = read_project_settings()
    deploy(settings)


if __name__ == "__main__":
    main()



class MissingFieldError(Exception):
    """Raised when fields are missing from mlops_settings.yaml"""


class FieldValueError(Exception):
    """Raised when field values are invalid"""


class MissingMetadataError(Exception):
    """Raised when fields are missing from the metadata (from mlops_settings.yaml)"""


class TooMuchMetadata(Exception):
    """Raised when there is too much metadata (from mlops_settings.yaml)"""


class TestError(Exception):
    """Raised when tests are failing"""


class DeploymentError(Exception):
    """Raised when deployment fails"""


class FunctionNameError(Exception):
    """Raised when function name (in CDF) is invalid"""


class FunctionCallError(Exception):
    """Raised when function call fails"""


class MissingClientError(Exception):
    """Raised when a global cognite client is missing, i.e. not set up"""


class VirtualEnvironmentError(Exception):
    """Raised when something is wrong when creating or deleting a virtual environment during deployments"""


class MLOpsError(Exception):
    """Generic MLOps exception"""


class MissingDeploymentFolder(Exception):
    """Raised when failing to download the deployment folder"""


class MissingResponseError(Exception):
    """Raised when function call return an empty response"""


class MissingModelArtifactsError(Exception):
    """Raised when queried model artifacts are missing"""


class MissingSettingsFileError(Exception):
    """Raised when mlops_settings.yaml is missing"""


class DependencyMismatchError(Exception):
    """Raised when there is a mismatch between main model and helper model dependencies"""



import os
import time
from typing import Any, Optional

from akerbp.mlops.core.logger_config import LOGGING_CONFIG

base_name = "main"


class CDFLogger:
    """
    Create a logger that mocks logging.Logger with only print statements.
    """

    def __init__(self, name: str = base_name, default_level: str = "INFO"):
        self.name = name
        self.datefmt = LOGGING_CONFIG["formatters"]["standard"]["datefmt"]
        self.level = default_level
        self.levels = {
            "DEBUG": 10,
            "INFO": 20,
            "WARNING": 30,
            "ERROR": 40,
            "CRITICAL": 50,
        }

    def _log(self, level: str, msg: str, args):
        """
        A helper method that prints the log message with the appropriate log level.
        """
        if self.levels[level] >= self.levels[self.level]:
            print(
                f"{time.strftime(self.datefmt)} - {level} - {self.name} - {msg % args}"
            )

    def info(self, msg: str, *args, **kwargs):
        self._log("INFO", msg, args)

    def debug(self, msg: str, *args, **kwargs):
        self._log("DEBUG", msg, args)

    def warning(self, msg: str, *args, **kwargs):
        self._log("WARNING", msg, args)

    def warn(self, msg: str, *args, **kwargs):
        self._log("WARNING", msg, args)

    def error(self, msg: str, *args, **kwargs):
        self._log("ERROR", msg, args)

    def critical(self, msg: str, *args, **kwargs):
        self._log("CRITICAL", msg, args)

    def exception(self, msg: str, *args, **kwargs):
        self._log("EXCEPTION", msg, args)

    def log(self, level: str, msg: str, *args, **kwargs):
        self._log(level, msg, args)

    def __repr__(self):
        return f"CDFLogger(name={self.name})"

    def __str__(self):
        return self.name


def get_logger(
    name: str = base_name,
    platform: Optional[str] = os.getenv("DEPLOYMENT_PLATFORM"),
) -> Any:  # Returns CDFLogger or logging.Logger but can't put import at top of module due to CDF issues
    """
    Set up a stream logger based on a global logging config.

    Args:
        name (str): name of the logger. Defaults to the global base_name variable

    Returns:
        (Logger): logger object
    """
    # can't use config module for this one or there would be circular imports
    if platform == "cdf":
        return CDFLogger(name)

    import logging
    from logging import config as logging_config

    logging_config.dictConfig(LOGGING_CONFIG)
    logger = logging.getLogger(name)
    logging.captureWarnings(True)

    return logger



# helpers.py
import importlib.util
import json
import os
import shutil
import subprocess
import sys
import tempfile
import venv
from importlib.metadata import PackageNotFoundError
from importlib.metadata import version as package_version
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, Union

from packaging.version import Version
from packaging.version import parse as parse_version

import akerbp.mlops as akerbp_mlops
from akerbp.mlops import __version__ as mlops_version
from akerbp.mlops.cdf import helpers as cdf
from akerbp.mlops.core.config import ENV_VARS, ServiceSettings
from akerbp.mlops.core.exceptions import (
    DeploymentError,
    TestError,
    VirtualEnvironmentError,
)
from akerbp.mlops.core.helpers import subprocess_wrapper
from akerbp.mlops.core.logger import get_logger
from akerbp.mlops.deployment import platforms

logger = get_logger(__name__)

envs = ENV_VARS
env = envs.model_env
service_name = envs.service_name
platform_methods = platforms.get_methods()


def is_unix() -> bool:
    """Checks whether the working OS is unix-based or not

    Returns:
        bool: True if os is unix-based
    """
    return os.name == "posix"


def get_repo_origin() -> str:
    """Get origin of the git repo

    Returns:
        (str): origin
    """
    origin = subprocess.check_output(
        ["git", "remote", "get-url", "--push", "origin"], encoding="UTF-8"
    ).rstrip()
    return origin


def replace_string_file(s_old: str, s_new: str, file: Path) -> None:
    """
    Replaces all occurrences of s_old with s_new in a specifyied file

    Args:
        s_old (str): old string
        s_new (str): new string
        file (Path): file to replace the string in
    """
    with file.open("r+") as f:
        data = f.read()
        if s_old not in data:
            logger.warning(f"Didn't find '{s_old}' in {file}")
        data = data.replace(s_old, s_new)
        new = os.linesep.join([s for s in data.splitlines() if s.strip()])
        f.seek(0)
        f.write(new)
        f.truncate()


def set_mlops_import(
    req_file: Path, platform: str, deployment_folder: Optional[str] = None
) -> None:
    """Set correct package version in requirements.txt

    Args:
        req_file (Path): path to requirements.txt for the model to deploy
        deployment_folder (Optional[str], optional): path to a deployment folder. Defaults to None.
            If this is set, the function assumes that the current version of the
            mlops package should be built and the version should be set to a wheel path.
    """
    if deployment_folder is not None:
        package_version = build_mlops_wheel(deployment_folder)
        set_to = f"./{package_version}[{platform}]"
        replace_string_file(
            "akerbp-mlops==MLOPS_VERSION",
            set_to,
            req_file,
        )
        logger.info(
            f"Set akerbp-mlops dependency to {package_version} in requirements.txt"
        )
    else:
        package_version = akerbp_mlops.__version__
        set_to = f"akerbp-mlops[{platform}]=={package_version}"
        replace_string_file(
            "akerbp-mlops==MLOPS_VERSION",
            set_to,
            req_file,
        )
        logger.info(f"Set akerbp-mlops==MLOPS_VERSION to {set_to} in requirements.txt")


def build_mlops_wheel(deployment_folder: str) -> str:
    """Build akerbp-mlops wheel in a virtual environment

    Warning:
        This function assumes that the current version of the mlops package
        is a cloned copy of the mlops repo. It will not work if the package
        is installed from PyPI!

    Note: This function assumes Poetry is installed and on path.

    Args:
        venv_dir (str): path to deployment folder where the wheel should be copied

    Returns:
        str: the wheel's filename.
    """
    # Get root of mlops repo
    root = Path(akerbp_mlops.__file__).parents[3]
    # Ensure the root contains pyproject.toml
    if not (root / "pyproject.toml").is_file():
        raise DeploymentError("Failed to find pyproject.toml in root")
    # Build wheel
    logger.info("Building akerbp-mlops wheel")
    subprocess_wrapper(["poetry", "build"], cwd=root)
    # Search for wheel name
    try:
        wheel_name = next((root / "dist").glob("*.whl"))
    except StopIteration as e:
        # Remove local build artifacts
        logger.info("Removing local build artifacts")
        shutil.rmtree(root / "dist")
        raise DeploymentError("Failed to build akerbp.mlops wheel") from e

    # Copy wheel to venv
    source = root / "dist" / wheel_name
    logger.info("Copying %s to virtual environment", source)
    shutil.copy(source, deployment_folder)

    # Remove local build artifacts
    logger.info("Removing local build artifacts")
    shutil.rmtree(root / "dist")

    return str(Path(wheel_name).resolve().name)


def to_folder(path: Path, folder_path: Path) -> None:
    """
    Copy folders, files or package data to a given folder.
    Note that if target exists it will be overwritten.

    Args:
        path: supported formats
            - file/folder path (Path): e,g, Path("my/folder")
            - module file (tuple/list): e.g. ("my.module", "my_file"). Module
            path has to be a string, but file name can be a Path object.
        folder_path (Path): folder to copy to
    """
    if isinstance(path, (tuple, list)):
        module_name, object_to_copy = path
        module_dir = Path(importlib.util.find_spec(module_name).origin).resolve().parent
        object_to_copy = str(object_to_copy)
        path_to_copy = module_dir / object_to_copy
        if path_to_copy.exists():
            if path_to_copy.is_dir():
                shutil.copytree(
                    path_to_copy,
                    folder_path / object_to_copy,
                    dirs_exist_ok=True,
                    ignore=shutil.ignore_patterns("__pycache__", ".pytest_cache"),
                )
            elif path_to_copy.is_file():
                shutil.copy(path_to_copy, folder_path)
        else:
            raise FileNotFoundError(
                f"Could not find {object_to_copy} in {module_name} directory"
            )
    elif path.is_dir():
        shutil.copytree(
            path,
            folder_path / path,
            dirs_exist_ok=True,
            ignore=shutil.ignore_patterns("__pycache__", ".pytest_cache"),
        )
    elif path.is_file():
        shutil.copy(path, folder_path)
    else:
        raise ValueError(f"{path} should be a file, folder or package resource")


def get_deployment_folder_content(
    deployment_folder: Path,
) -> Dict:
    """Returns the content of the deployment folder as a dictionary with keys being
    directories/subdirectories, and values the corresponding content.
    If venv_dir is passed as an argument we ignore the content of the corresponding virtual environment

    Args:
        deployment_folder (Path): path to deployment_folderfolder

    Returns:
        Dict: content of deployment folder
    """
    dirwalk = os.walk(deployment_folder)
    content = {}
    for root, dirs, files in dirwalk:
        if "__pycache__" in root.split("/"):
            continue
        if ".pytest_cache" in root.split("/"):
            continue
        dirs = [d for d in dirs if d not in ["__pycache__", ".pytest_cache"]]
        content[root] = files
        if len(dirs) > 0:
            content[root].extend(dirs)
            for subdir in dirs:
                content[os.path.join(root, subdir)] = files  # noqa: PTH118
        else:
            content[root] = files
    return content


def copy_to_deployment_folder(lst: Dict, deployment_folder: Path) -> None:
    """
    Copy a list of files/folders to a deployment folder

    Args:
        lst (dict): key is the nickname of the file/folder (used for
        logger) and the value is the path (see `to_folder` for supported
        formats)
        deployment_folder (Path): Path object for the deployment folder

    """
    for k, v in lst.items():
        if v:
            logger.debug(f"{k} => deployment folder")
            to_folder(v, deployment_folder)
        else:
            logger.warning(f"{k} has no value")


def update_pip(venv_path: Path, **kwargs) -> None:
    is_unix_os = kwargs.get("is_unix_os", True)
    setup_venv = kwargs.get("setup_venv", True)
    old_executable = sys.executable
    if setup_venv:
        if is_unix_os:
            sys.executable = str(venv_path / "bin" / "python")
            c = [
                sys.executable,
                "-m",
                "pip",
                "install",
                "--upgrade",
                "pip",
            ]
        else:
            sys.executable = str(venv_path / "Scripts" / "python")
            c = [
                sys.executable,
                "-m",
                "pip",
                "install",
                "--upgrade",
                "pip",
            ]
    else:
        c = ["python", "-m", "pip", "install", "--upgrade", "pip"]
    logger.info("Updating pip")
    subprocess_wrapper(c)
    sys.executable = old_executable


def install_requirements(req_file: str, venv_path: Path, **kwargs) -> None:
    """install model requirements

    Args:
        req_file (str): path to requirements file
    """
    requirement_file = Path(req_file).resolve()
    with_deps = kwargs.get("with_deps", True)
    setup_venv = kwargs.get("setup_venv", False)
    is_unix_os = is_unix()
    update_pip(
        venv_path=venv_path,
        is_unix_os=is_unix_os,
        setup_venv=setup_venv,
    )
    logger.info(f"Installing python requirement file {req_file}:")
    # Cat file contents to log
    with requirement_file.open() as f:
        req_contents = f.read()
        logger.info(req_contents)
    requirements = list(filter(None, req_contents.splitlines()))
    install_special_mlpet = False
    if "akerbp-mlpet" in req_contents:
        new_requirements = []
        for requirement in requirements:
            if "akerbp-mlpet" in requirement:
                mlpet_version = parse_version(requirement.split("==")[1])
                if sys.version_info >= (3, 10) and mlpet_version < Version("4.0.0"):
                    install_special_mlpet = True
                    logger.warning(
                        "akerbp-mlpet is specified in requirements.txt. Treating this package specially due to python version constraints"
                    )
                    mlpet_version = str(mlpet_version)  # type: ignore
                else:
                    new_requirements.append(requirement)
            else:
                new_requirements.append(requirement)
        requirements = new_requirements
    install_special_scikit_learn = False
    if "scikit-learn" in req_contents and sys.version_info >= (3, 10):
        new_requirements = []
        for requirement in requirements:
            if "scikit-learn" in requirement:
                sklearn_version = parse_version(requirement.split("==")[1])
                # If scikit-learn version is less than 1.2, we need to use Cython<3 and install it without build isolation
                # ref scikit-learn GH issue #26858
                if sklearn_version < Version("1.2.0"):
                    logger.warning(
                        "scikit-learn version is less than 1.2. Installing Cython<3 and scikit-learn without build isolation"
                    )
                    new_requirements.extend([
                        "cython==0.29.36",
                        "wheel",
                        "setuptools<60",
                    ])
                    install_special_scikit_learn = True
                    sklearn_version = str(sklearn_version)  # type: ignore
                else:
                    new_requirements.append(requirement)
            else:
                new_requirements.append(requirement)
        requirements = new_requirements
    old_executable = sys.executable
    if with_deps:
        if setup_venv:
            if is_unix_os:
                sys.executable = str(venv_path / "bin" / "python")
                c = [
                    sys.executable,
                    "-m",
                    "pip",
                    "install",
                ]
            else:
                sys.executable = str(venv_path / "Scripts" / "python")
                c = [
                    sys.executable,
                    "-m",
                    "pip",
                    "install",
                ]
        else:
            c = ["pip", "install"]
    else:
        if setup_venv:
            if is_unix_os:
                sys.executable = str(venv_path / "bin" / "python")
                c = [
                    sys.executable,
                    "-m",
                    "pip",
                    "install",
                    "--no-deps",
                ]
            else:
                sys.executable = str(venv_path / "Scripts" / "python")
                c = [
                    sys.executable,
                    "-m",
                    "pip",
                    "install",
                ]
        else:
            c = ["pip", "install", "--no-deps"]
    install_command = c + requirements
    subprocess_wrapper(install_command)
    if install_special_mlpet:
        subprocess_wrapper(
            c + ["--ignore-requires-python", "akerbp-mlpet==" + mlpet_version]  # type: ignore
        )
    if install_special_scikit_learn:
        subprocess_wrapper(
            c
            + [
                "scikit-learn==" + sklearn_version,  # type: ignore
                "--no-build-isolation",
            ]
        )
    sys.executable = old_executable


def create_venv(venv_name: str) -> Path:
    """
    Create a virtual environment with the given name. The virtual environment
    is created in the current working directory from where this function is called.

    Args:
        venv_name (str): The name of the virtual environment.

    Returns:
        Path: The path to the created virtual environment.

    Raises:
        VirtualEnvironmentError: If the virtual environment was not created successfully.
    """
    venv_path = (Path.cwd() / venv_name).resolve()
    logger.info(f"Creating virtual environment {venv_name} at {venv_path=}")
    venv.create(venv_path, with_pip=True)
    if venv_path.is_dir():
        logger.info(f"Successfully created virtual environment {venv_name}")
    else:
        raise VirtualEnvironmentError("Virtual environment was not created")
    return venv_path


def delete_venv(venv_path: Path) -> None:
    logger.info(f"Deleting virtual environment {venv_path}")
    try:
        shutil.rmtree(venv_path.resolve())
    except Exception as e:
        raise VirtualEnvironmentError(
            f"Failed to delete virtual environment {venv_path}"
        ) from e
    logger.info(f"Virtual environment {venv_path} sucsessfully deleted")


def set_up_requirements(c: ServiceSettings, **kwargs) -> Path:
    """
    Set up a "requirements.txt" file at the top of the deployment folder
    (assumed to be the current directory), update config and install
    dependencies (unless in dev)

    Args:
        c (ServiceSettings): service settings a specified in the config file

    Keyword Args:
        install (bool): Whether to install the dependencies, defaults to True
    """
    logger.info("Create requirement file")
    install_reqs = kwargs.get("install", True)
    with_deps = kwargs.get("with_deps", True)
    venv_name = kwargs.get("venv_name", "mlops-venv")
    setup_venv = kwargs.get("setup_venv", False)
    unit_testing = kwargs.get("unit_testing", False)
    if setup_venv:
        venv_path = create_venv(venv_name=venv_name)
    else:
        venv_path = Path.cwd()

    if not unit_testing:
        shutil.copyfile(c.req_file, "requirements.txt")
        c.req_file = Path("requirements.txt").resolve()

    if env != "dev" or install_reqs:
        install_requirements(
            c.req_file, venv_path=venv_path, with_deps=with_deps, setup_venv=setup_venv
        )
    else:
        logger.info("Skipping installation of requirements.txt")

    return venv_path


def deployment_folder_path(model: str) -> Path:
    """Generate path to deployment folder, which is on the form "mlops_<model>"

    Args:
        model (str): model name

    Returns:
        Path: path to the deployment folder
    """
    return Path(f"mlops_{model}")


def rm_deployment_folder(model: str) -> None:
    logger.info("Deleting deployment folder")
    deployment_folder = deployment_folder_path(model)
    if deployment_folder.exists():
        shutil.rmtree(deployment_folder)


def run_tests(c: ServiceSettings, **kwargs) -> Any:
    """Helper function that runs unit tests and returns a test payload
    if run during deployment

    Args:
        c (ServiceSettings): Service settings object containing model settings

    Keyword Args:
        setup_venv (bool): whether to setup an ephemeral virtual environment.
            Defaults to False
    Raises:
        TestError: If unit tests are failing

    Returns:
        Union[Dict[str, Any], None]: Return test payload if test file is specified in 'mlops_settings.yaml'
    """
    if c.test_file:
        deploy = kwargs.get("deploy", True)
        setup_venv = kwargs.get("setup_venv", False)
        executable = None
        if setup_venv:
            logger.info("Setting up ephemeral virtual environment")
            venv_path = set_up_requirements(
                c,
                install=True,
                setup_venv=setup_venv,
            )
            if is_unix():
                executable = str(venv_path / "bin" / "python")
            else:
                executable = str(venv_path / "Scripts" / "python")
        else:
            set_up_requirements(c, install=True)
        test_command = [
            sys.executable if executable is None else executable,
            "-m",
            "akerbp.mlops.services.test_service",
        ]
        logger.info(f"Running tests for model {c.model_name}")
        failed_test = False
        try:
            output = subprocess_wrapper(
                test_command,
                skiplines=[r'{"input": \['],  # skip payload lines
                encoding="UTF-8",
            )
        except subprocess.CalledProcessError as e:
            output = e.output
            failed_test = True
        model_input = None
        for log_line in output.splitlines():  # type: ignore
            log_line = log_line.strip()
            if "secrets =" in log_line:
                continue  # Don't print secrets
            if log_line.startswith('{"input": ['):
                # Extract payload for downstream testing of deployed model
                model_input = json.loads(log_line)
                if deploy:
                    logger.info(
                        "Payload for downstream testing of deployed model obtained"
                    )
        if failed_test:
            raise TestError("Unit tests failed :( See the above traceback")
        if model_input is None and deploy:
            raise TestError(
                "Test was not able to extract the payload for downstream testing of deployed model"
            )
        logger.info("Unit tests passed :)")
        if setup_venv:
            delete_venv(venv_path=venv_path)
            logger.info("Ephemeral virtual environment deleted")
        return model_input
    else:
        logger.warning(
            "No test file specified in 'mlops_settings.yaml', skipping tests"
        )
        return {}


def do_deploy(
    c: ServiceSettings,
    env: str,
    service_name: str,
    function_name: str,
    deployment_folder: str,
    platform_methods: Dict = platform_methods,
    **kwargs,
) -> None:
    external_id = function_name
    platform = c.platform
    python_runtime = c.python_version
    deploy_function, _, test_function, metadata_validation_function = platform_methods[
        platform
    ]
    logger.info(f"Starting deployment of model {c.model_name} to {env}")
    # Get the latest artifact version uploaded to CDF Files and tag the model metadata with this version number
    if platform == "cdf":
        if c.artifact_version is None:
            latest_artifact_version = cdf.get_latest_artifact_version(
                external_id=external_id
            )
            logger.info(
                f"Latest artifact version in {env} is {latest_artifact_version}"
            )
            artifact_version = latest_artifact_version
        else:
            artifact_version = c.artifact_version
    elif platform == "gc":
        pass
    else:
        raise NotImplementedError(
            f"Functionality for deploying models to platform {platform} is not implemented!"
        )

    # Tag model metadata with version number, mlops version and mlpet version if present
    model_info = c.info[service_name]
    model_info["metadata"]["version"] = str(artifact_version)
    model_info["metadata"]["akerbp-mlops_version"] = mlops_version
    try:
        mlpet_version = package_version("akerbp-mlpet")
    except PackageNotFoundError:
        mlpet_version = "not found"
    model_info["metadata"]["akerbp-mlpet_version"] = mlpet_version

    # Validate model metadata
    logger.info(f"Validating model metadata for model {c.model_name}")
    try:
        metadata_validation_function(model_info["metadata"])
    except Exception as e:
        raise DeploymentError(f"Deployment failed with message: \n{str(e)}") from e

    # Run unit tests and get test payload before deploying
    logger.info(
        f"Running tests for model {c.human_friendly_model_name} before deploying to {env}"
    )
    setup_venv = kwargs.pop("setup_venv", False)
    if setup_venv:
        test_payload = run_tests(c, setup_venv=setup_venv)
    else:
        test_payload = run_tests(c, setup_venv=setup_venv)

    # Get content of deployment folder
    deployment_folder_content = get_deployment_folder_content(deployment_folder=Path())
    logger.info(
        f"Deployment folder '{deployment_folder}' now contains the following: {deployment_folder_content}"
    )
    logger.info(
        f"Deploying function {c.human_friendly_model_name} with external id {external_id} to {platform}"
    )
    try:
        function = deploy_function(
            c.human_friendly_model_name,
            external_id,
            info=model_info,
            python_runtime=python_runtime,
        )
    except Exception as e:
        raise DeploymentError(f"Deployment failed with message: \n{str(e)}") from e
    if c.test_file:
        logger.info(f"Make a test call to function with external id {external_id}")
        try:
            test_function(external_id, test_payload)
        except Exception as e:
            raise TestError(
                f"Test of deployed model failed with message: \n{str(e)}"
            ) from e
    else:
        logger.warning(
            f"No test file was set up. End-to-end test skipped for function {external_id}"
        )

    if platform == "cdf" and env == "prod":
        # Create a schedule for keeping the latest function warm in prod
        logger.info(
            f"Creating a schedule for keeping the function {external_id} warm on weekdays during extended working hours"
        )
        cdf.setup_schedule_for_latest_model_in_prod(
            external_id=function_name, function_id=function.id
        )
        # Redeploy latest function with a predictable external id (model-service-env)
        numbered_external_id = function_name + "-" + str(latest_artifact_version)
        logger.info(
            f"Redeploying numbered model {c.human_friendly_model_name} with external id {external_id} to {platform} in {env}"
        )
        redeploy_model_with_numbered_external_id(
            c,
            numbered_external_id=numbered_external_id,
            test_payload=test_payload,
            info=model_info,
        )

    if platform == "cdf":
        for payload in test_payload["input"]:
            test_payload_file_external_id = payload.get("input_file_reference", None)
            if test_payload_file_external_id is not None:
                logger.info("Deleting test payload uploaded to CDF Files")
                cdf.delete_test_payload(external_id=test_payload_file_external_id)
        logger.info("Initiating garbage collection of old models in CDF")
        cdf.garbage_collection(
            c,
            function_name,
            env,
            remove_artifacts=c.model_name == "mlopsdemo",
        )


def redeploy_model_with_numbered_external_id(
    c: ServiceSettings,
    numbered_external_id: str,
    test_payload: Any,
    info: Dict[str, Union[str, Dict[str, str]]],
    platform_methods: Dict = platform_methods,
) -> None:
    deploy_function, _, test_function, _ = platform_methods[c.platform]

    try:
        deploy_function(
            c.human_friendly_model_name,
            numbered_external_id,
            info=info,
            python_runtime=c.python_version,
        )

    except Exception as e:
        raise DeploymentError(
            f"Redeployment of numbered model failed with message:  \n{str(e)}"
        ) from e

    if c.test_file:
        try:
            test_function(numbered_external_id, test_payload)
        except Exception as e:
            raise TestError(
                f"Testing the newly redeployed latest model failed with message: \n{str(e)}"
            ) from e
    else:
        logger.warning("No test file was specified in the settings, skipping tests")


def create_temporary_copy(path: Path) -> Path:
    logger.info("Creating a temporary copy of %s", path)
    _, temp_path = tempfile.mkstemp()
    shutil.copy2(path, temp_path)
    return Path(temp_path).resolve()


def requirements_to_dict(path_to_requirements: Path) -> Dict[str, str]:
    """Convert a requirements file to a dictionary

    Args:
        path_to_requirements (Path): path to requirements file

    Returns:
        Dict[str, str]: dictionary with package name as key and version as value
    """
    requirements = {}
    with path_to_requirements.open() as f:
        for line in f.readlines():
            if "==" in line:
                package, version = line.split("==")
                requirements[package] = version.strip()
    return requirements


def check_mismatch_in_model_requirements(
    main_model_requirements: Dict[str, str],
    helper_model_requirements: Dict[str, str],
    packages_to_check: List[str],
) -> Tuple[bool, List[str]]:
    """Check if there is a mismatch between main and helper model requirements for a given list of packages
    Returns True if there is a mismatch in at least one of the listed packages, False otherwise

    Args:
        main_model_requirements (Dict[str, str]): _description_
        helper_model_requirements (Dict[str, str]): _description_

    Returns:
        bool: Whether there is a mismatch between main and helper model requirements
    """
    mismatched_requirements = []
    mismatched_packages = []
    helper_model_requirements_ = {
        k.replace(".", "-"): v for k, v in helper_model_requirements.items()
    }
    for key in main_model_requirements.keys():
        if key in helper_model_requirements_.keys() and key in packages_to_check:
            if main_model_requirements[key] != helper_model_requirements_[key]:
                mismatched_requirements.append(True)
                mismatched_packages.append(key)
            else:
                mismatched_requirements.append(False)

    if sum(mismatched_requirements) > 0:
        return True, mismatched_packages
    else:
        return False, []



"""
service.py

Training service
"""

# type: ignore
from importlib import import_module
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Any, Callable, Dict

import akerbp.mlops.model_manager as mm
from akerbp.mlops.core import config
from akerbp.mlops.core.logger import get_logger

logger = get_logger(__name__)

c = config.read_service_settings()
env = config.ENV_VARS.model_env

model_module = import_module(c.model_import_path)
train = model_module.train
ModelException = model_module.ModelException


def _saver(path: Path, metadata: Dict, client_secrets: Dict) -> Any:
    """
    Upload model folder to CDF Files
    """
    mm.setup(client_secrets, c.dataset)
    model_info = mm.upload_new_model_version(c.model_name, env, path, metadata)
    return model_info


def service(data: Dict, secrets: Dict, saver: Callable = _saver) -> Dict:
    """
    Training service
    Inputs:
        - data: dictionary, data passed by the user through the API
        - secrets: dictionary with api keys
        - saver: an object that saves the model folder

    Output:
        - Dictionary with status 'ok' or 'error' as keys.
            status == 'ok'    -> there is a 'training' key as well
                                (data on the model file)
            status == 'error' -> there is a 'message' key as well
    """
    try:
        with TemporaryDirectory() as temp_dir:
            metadata = train(data=data, folder_path=temp_dir, secrets=secrets)
            logger.debug(f"{metadata=}")
            model_info = saver(temp_dir, metadata, secrets)
        return {"status": "ok", "training": model_info}
    except ModelException as e:
        error_message = f"Training failed. Message: {e}"
        logger.error(error_message)
        return {"status": "error", "message": error_message}



# helpers.py
import os
import re
import subprocess
import time
import traceback
from pathlib import Path
from typing import Dict, List, Optional, Union

import requests

from akerbp.mlops.core import mappings
from akerbp.mlops.core.logger import get_logger

logger = get_logger(__name__)


def get_top_folder(path: Union[Path, str]) -> Union[Path, str]:
    """
    Get the top folder of a path.

    Args:
        path (Union[Path, str]): Path to get the top folder from.

    Returns:
        (Union[Path, str]): top parent folder in path
    """
    if isinstance(path, str):
        return Path(path).parts[0]
    elif isinstance(path, Path):
        return path.parents[len(path.parents) - 2]


def as_import_path(file_path: Optional[str]) -> Optional[str]:
    """Return path as an import path for python modules


    Args:
        file_path (str, optional): path to file

    Returns:
        (Union[str, None]): path to .py-file with .py extension removed(for importing modules)
    """
    if file_path:
        if not isinstance(file_path, str):
            file_path = str(file_path)
        return file_path.replace(os.sep, ".").replace(".py", "")
    else:
        logger.debug("Empty file path -> empty import path returned")
        return None


def confirm_prompt(question: str) -> bool:
    """Helper function to ask user for confirmation

    Args:
        question (str): question to ask user

    Returns:
        (bool): True if user confirms, False otherwise
    """
    reply = None
    while reply not in ("y", "n"):
        reply = input(f"{question} (y/n): ").lower()
    return reply == "y"


def input_data_validation(
    required_input: List[str],
    input_payload: Dict,
) -> bool:
    """Check if the input data contains all the required input curves.
    Note that the input potentially consists of multiple wells according to the payload structures.

    If input logs are provided in the payload, they need to be provided in the field 'input_logs',
    otherwise the validation fails.
    If no input logs are provided, the validation passes if either the field 'well' or
    'well' and 'keyword_arguments' are specified

    Args:
        required_input (List[str]): list of required input curves
        input (Dict): dictionary containing the input data

    Returns:
        bool: whether the input data contains all the required input curves or not
    """
    if required_input is None:
        print("No required input here")
        return True
    curve_mapping = mappings.curve_mappings
    validated = []
    input_wells = input_payload["input"]
    for well_data in input_wells:
        try:
            curves_in_well_data = list(well_data["input_logs"].keys())
            curves_in_well_data_standardized = [
                curve_mapping[curve]
                if curve != "WELL" and curve in curve_mapping.keys()
                else curve
                for curve in curves_in_well_data
            ]
            diff = set(required_input) - set(curves_in_well_data_standardized)

            if len(diff) > 0:
                logger.warning(
                    f"Not all required input curves are provided. Missing: {diff}"
                )
                validated.append(False)
            else:
                validated.append(True)
        except KeyError:
            wrong_input_log_specification = False
            for key in well_data.keys():
                if isinstance(well_data[key], dict) and len(well_data[key].keys()) > 0:
                    wrong_input_log_specification = True
                    break
            if wrong_input_log_specification:
                logger.warning(
                    f"Data is required to be specified as 'input_logs' in the payload, not '{key}'"
                )
                validated.append(False)
            elif "well" in well_data.keys() and "keyword_arguments" in well_data.keys():
                validated.append(True)
            elif "well" in well_data.keys():
                validated.append(True)
            else:
                logger.warning(
                    "Payload not properly specified. Need to specify field 'well', and optionally 'keyword_arguments' if no input logs are provided"
                )
    if sum(validated) < len(input_wells):
        return False
    else:
        return True


def response_validation(response: List[Dict]) -> bool:
    """Validates the reponse of each model call,
    making sure that it contains at least 'well_name' and 'DEPTH'

    Args:
        response (List[Dict]): response from predict function
        required_output (List[str]): List of required output curves

    Returns:
        bool: whether or not response is valid in terms of structure and required output curves
    """
    required_output_fields = {"well_name", "DEPTH"}
    validated_response = []
    for res in response:
        response_output_curves = set(res.keys())
        if required_output_fields.issubset(response_output_curves):
            validated_response.append(True)
        else:
            logger.warning(
                f"The response does not have the required fields: {required_output_fields}"
            )
            validated_response.append(False)
    return sum(validated_response) == len(validated_response)


def subprocess_wrapper(
    cmd: List[str], retry: int = 1, log_output=True, skiplines=None, **kwargs
) -> str:
    encoding = kwargs.pop("encoding", "utf-8")
    if skiplines is None:
        pat = None
    else:
        pat = re.compile("|".join(skiplines))
    output = ""
    for i in range(retry):
        completed_process = False
        try:
            logger.info(
                f"Running subprocess.check_output with the following commands: {cmd}"
            )
            output = subprocess.check_output(cmd, encoding=encoding, **kwargs)
        except subprocess.CalledProcessError as e:
            output = e.output
            if i == (retry - 1):
                raise e
            else:
                logger.warning(
                    f"Command {cmd} failed with exit code {e.returncode}. Retrying..."
                )
                time.sleep(15)
        else:
            completed_process = True
        finally:
            if log_output:
                for line in output.splitlines():
                    line.strip()
                    if pat is not None and pat.search(line):
                        logger.info(
                            f"Skipping subprocess output due to provided skiplines pattern {pat}"
                        )
                        continue
                    if completed_process:
                        logger.info(line)
                    else:
                        logger.error(line)
    return output


def requests_wrapper(
    request_type: str,
    url: str,
    data: Optional[Dict] = None,
    headers: Optional[Dict] = None,
    **kwargs,
) -> requests.Response:
    """
    Wrapper for requests.get and requests.post with error handling

    Args:
        request_type (str): type of request, either 'get' or 'post'
        url (str): url to send request to
        data (Dict, optional): data to send in request. Defaults to None.
        headers (Dict, optional): headers to send in request. Defaults to None.

    Returns:
        (requests.Response): response from request
    """
    try:
        response = requests.request(
            request_type,
            url,
            data=data,
            headers=headers,
            **kwargs,
        )
        response.raise_for_status()
    except requests.exceptions.HTTPError as e:
        trace = traceback.format_exc()
        logger.error(
            f"Something failed with your request to the following {url}: {e}\n{trace}"
        )
        raise e
    except requests.exceptions.RequestException as e:
        trace = traceback.format_exc()
        logger.error(f"Error when sending request to {url}: {e}\n{trace}")
        raise e
    return response



# This can be run after package has been installed
def main() -> None:
    import os

    os.environ["SERVICE_NAME"] = "training"
    import akerbp.mlops.deployment.deploy as deploy

    deploy.main()


if __name__ == "__main__":
    main()



curve_mappings = {
    "4A_PRES": "4A_PRESS",
    "A16L": "RSHA",
    "A28H": "RMED",
    "A40L": "RDEP",
    "AC": "AC",
    "ACS": "ACS",
    "AI": "AI",
    "AI10": "RSHA",
    "AI60": "RMED",
    "AI90": "RDEP",
    "AO10": "RSHA",
    "AO60": "RMED",
    "AO90": "RDEP",
    "APLC": "APLC",
    "AT10": "RSHA",
    "AT60": "RMED",
    "AT90": "RDEP",
    "BFV": "BFV",
    "BFV_UHR": "BFV",
    "BPHI": "BPHI",
    "BS": "BS",
    "CAL": "CALI",
    "CALI": "CALI",
    "CALS": "CALI",
    "CGR": "CGR",
    "CMRT": "MRP",
    "CNC": "NEU",
    "DCAV": "CALI",
    "DCOR": "DENC",
    "DEN": "DEN",
    "DENC": "DENC",
    "DEN_VIRGIN": "DEN_VIRGIN",
    "DRHO": "DENC",
    "DT": "AC",
    "DTC": "AC",
    "DTCO": "AC",
    "DTS": "ACS",
    "DTSM": "ACS",
    "FFV": "FFV",
    "FFV_UHR": "FFV",
    "FPRESS": "FPRESS",
    "GR": "GR",
    "GRMA": "GR",
    "GR_EDTC": "GR",
    "HAC": "AC",
    "HCAL": "CALI",
    "HCGR": "CGR",
    "HCNC": "NEU",
    "HCNL": "NEU",
    "HCOR": "DENC",
    "HCORR": "DENC",
    "HDAR": "CALI",
    "HDEN": "DEN",
    "HDRA": "DENC",
    "HDRH": "DENC",
    "HDT": "AC",
    "HFK": "K",
    "HGR": "GR",
    "HPHI": "NEU",
    "HRD": "RDEP",
    "HRHO": "DEN",
    "HRM": "RMED",
    "HRS": "RMIC",
    "HSGR": "SGR",
    "HTHO": "TH",
    "HURA": "U",
    "IDDR": "DENC",
    "IDPE": "PEF",
    "IDRO": "DEN",
    "ILD": "RDEP",
    "ILM": "RMED",
    "K": "K",
    "LLD": "RDEP",
    "LLS": "RSHA",
    "M": "M",
    "MRP": "MRP",
    "MSFL": "RMIC",
    "NEU": "NEU",
    "NPHI": "NEU",
    "NPOR": "NEU",
    "P16L": "RSHA",
    "P28L": "RMED",
    "P40H": "RDEP",
    "P40L": "RDEP",
    "PE": "PEF",
    "PEB": "PEF",
    "PEF": "PEF",
    "PEFZ": "PEF",
    "POTA": "K",
    "RD": "RDEP",
    "RDEP": "RDEP",
    "RH72_1DF": "RH",
    "RHGA": "RHGE",
    "RHGE": "RHGE",
    "RHO8": "DEN",
    "RHOB": "DEN",
    "RHOZ": "DEN",
    "RMED": "RMED",
    "RMIC": "RMIC",
    "RMLL": "RMIC",
    "ROBB": "DEN",
    "ROP": "ROP",
    "RS": "RSHA",
    "RSHA": "RSHA",
    "RT": "RT",
    "RT_HRLT": "RDEP",
    "RV72_1DF": "RV",
    "RXO": "RXO",
    "RXOZ": "RXO",
    "RXO_ADT": "RXO",
    "RXO_HRLT": "RXO",
    "SFLU": "SFLU",
    "SGR": "SGR",
    "SHAR": "RSHA",
    "SP": "SP",
    "SPBR": "SP",
    "TEMP": "TEMP",
    "TH": "TH",
    "THOR": "TH",
    "TNPH": "NEU",
    "U": "U",
    "UCAV": "CALI",
    "URAN": "U",
    "WANH": "WANH",
    "WCAR": "WCAR",
    "WCLA": "WCLA",
    "WCLC": "WCLC",
    "WCOA": "WCOA",
    "WDOL": "WDOL",
    "WEVA": "WEVA",
    "WPYR": "WPYR",
    "WQFM": "WQFM",
    "WSID": "WSID",
    "ZCOR": "DENC",
    "ZDEN": "DEN",
    "ZONE": "lsuName",
    "lsuName": "lsuName",
    "wellName": "well_name",
    "well_name": "well_name",
    "WELL": "well_name",
    "well": "well_name",
}
